import abc
import difflib
import os
import re
import sys
import time
from collections import OrderedDict as odict
from itertools import groupby
from operator import itemgetter
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

import colorama
from annet.annlib import jsontools
from annet.annlib.netdev.views.hardware import HardwareView
from annet.annlib.rbparser.platform import VENDOR_REVERSES
from annet.annlib.types import GeneratorType
from contextlog import get_logger

import annet.deploy
from annet import cli_args
from annet import diff as ann_diff
from annet import filtering
from annet import gen as ann_gen
from annet import patching, rulebook, tabparser, tracing
from annet.hardware import hardware_connector
from annet.output import (
    LABEL_NEW_PREFIX,
    format_file_diff,
    output_driver_connector,
    print_err_label,
)
from annet.parallel import Parallel, TaskResult
from annet.reference import RefTracker
from annet.storage import Device, Storage, storage_connector
from annet.types import Diff, ExitCode, OldNewResult, Op, PCDiff, PCDiffFile


live_configs = ann_gen.live_configs

DEFAULT_INDENT = "  "


def patch_from_pre(pre, hw, rb, add_comments, ref_track=None, do_commit=True):
    if not ref_track:
        ref_track = RefTracker()
    orderer = patching.Orderer(rb["ordering"], hw.vendor)
    orderer.ref_insert(ref_track)
    return patching.make_patch(
        pre=pre,
        rb=rb,
        hw=hw,
        add_comments=add_comments,
        orderer=orderer,
        do_commit=do_commit,
    )


def _diff_and_patch(
    device, old, new, acl_rules, filter_acl_rules,
    add_comments, ref_track=None, do_commit=True, rb=None
) -> Tuple[Diff, Dict]:
    if rb is None:
        rb = rulebook.get_rulebook(device.hw)
    # [NOCDEV-5532] Передаем в diff только релевантные для logic'a части конфига
    if acl_rules is not None:
        old = patching.apply_acl(old, acl_rules)
        new = patching.apply_acl(new, acl_rules, with_annotations=add_comments)

    diff_tree = patching.make_diff(old, new, rb, [acl_rules, filter_acl_rules])
    pre = patching.make_pre(diff_tree)
    patch_tree = patch_from_pre(pre, device.hw, rb, add_comments, ref_track, do_commit)
    diff_tree = patching.strip_unchanged(diff_tree)

    return (diff_tree, patch_tree)


# =====
def _read_old_new_diff_patch(old: Dict[str, Dict], new: Dict[str, Dict], hw: HardwareView, add_comments: bool):
    rb = rulebook.get_rulebook(hw)
    diff_obj = patching.make_diff(old, new, rb, [])
    diff_obj = patching.strip_unchanged(diff_obj)
    pre = patching.make_pre(diff_obj)
    patchtree = patch_from_pre(pre, hw, rb, add_comments)
    return rb, diff_obj, pre, patchtree


def _read_old_new_hw(old_path: str, new_path: str, args: cli_args.FileInputOptions):
    _logger = get_logger()
    hw = args.hw
    if isinstance(args.hw, str):
        hw = HardwareView(args.hw, "")

    old, old_hw, old_score = _read_device_config(old_path, hw)
    new, new_hw, new_score = _read_device_config(new_path, hw)
    hw = new_hw
    if old_score > new_score:
        hw = old_hw
    if old_hw != new_hw:
        _logger.debug("Old and new detected hw differs, assume %r", hw)
    dest_name = os.path.basename(new_path)
    return dest_name, old, new, hw


@tracing.function
def _read_old_new_cfgdumps(args: cli_args.FileInputOptions):
    _logger = get_logger()
    old_path, new_path = os.path.normpath(args.old), os.path.normpath(args.new)
    if not os.path.isdir(old_path):
        yield (old_path, new_path)
        return
    _logger.info("Scanning cfgdumps: %s/*.cfg ...", old_path)
    cfgdump_reg = re.compile(r"^[^\s]+\.cfg$")
    if os.path.isdir(old_path) and os.path.isdir(new_path):
        if cfgdump_reg.match(os.path.basename(old_path)) and cfgdump_reg.match(os.path.basename(new_path)):
            yield (old_path, new_path)
    for name in os.listdir(old_path):
        old_path_name = os.path.join(old_path, name)
        new_path_name = os.path.join(new_path, name)
        if not os.path.exists(new_path_name):
            _logger.debug("Ignoring file %s: not exist %s", name, new_path_name)
            continue
        yield (old_path_name, new_path_name)


def _read_device_config(path, hw):
    _logger = get_logger()
    _logger.debug("Reading %r ...", path)
    score = 1

    with open(path) as cfgdump_file:
        text = cfgdump_file.read()
    try:
        if not hw:
            hw, score = guess_hw(text)
        config = tabparser.parse_to_tree(
            text=text,
            splitter=tabparser.make_formatter(hw).split,
        )
        return config, hw, score
    except tabparser.ParserError:
        _logger.exception("Parser error: %r", path)
        raise


# =====
def _format_patch_blocks(patch_tree, hw, indent):
    formatter = tabparser.make_formatter(hw, indent=indent)
    return formatter.patch(patch_tree)


# =====
def _print_pre_as_diff(pre, show_rules, indent, file=None, _level=0):
    for (raw_rule, content) in sorted(pre.items(), key=itemgetter(0)):
        rule_printed = False
        for (op, sign) in [  # FIXME: Not very effective
            (Op.REMOVED, colorama.Fore.RED + "-"),
            (Op.ADDED, colorama.Fore.GREEN + "+"),
            (Op.AFFECTED, colorama.Fore.CYAN + " "),
        ]:
            items = content["items"].items()
            if not content["attrs"]["multiline"]:
                items = sorted(items, key=itemgetter(0))
            for (_, diff) in items:  # pylint: disable=redefined-outer-name
                if show_rules and not rule_printed and not raw_rule == "__MULTILINE_BODY__":
                    print("%s%s# %s%s%s" % (colorama.Style.BRIGHT, colorama.Fore.BLACK, (indent * _level),
                                            raw_rule, colorama.Style.RESET_ALL), file=file)
                    rule_printed = True
                for item in sorted(diff[op], key=itemgetter("row")):
                    print("%s%s%s %s%s" % (colorama.Style.BRIGHT, sign, (indent * _level),
                                           item["row"], colorama.Style.RESET_ALL), file=file)
                    if len(item["children"]) != 0:
                        _print_pre_as_diff(item["children"], show_rules, indent, file, _level + 1)
                        rule_printed = False


def log_host_progress_cb(pool: Parallel, task_result: TaskResult):
    progress_logger = get_logger("progress")
    args = cast(cli_args.QueryOptions, pool.args[0])
    connector = storage_connector.get()
    storage_opts = connector.opts().from_cli_opts(args)
    with connector.storage()(storage_opts) as storage:
        hosts = storage.resolve_fdnds_by_query(args.query)
    perc = int(pool.tasks_done / len(hosts) * 100)
    fqdn = hosts[task_result.device_id]
    elapsed_time = "%dsec" % int(time.monotonic() - task_result.extra["start_time"])
    if task_result.extra.get("regression", False):
        status = task_result.extra["status"]
        status_color = task_result.extra["status_color"]
        message = task_result.extra["message"]
    else:
        status = "OK" if task_result.exc is None else "FAIL"
        status_color = colorama.Fore.GREEN if status == "OK" else colorama.Fore.RED
        message = "" if status == "OK" else str(task_result.exc)
    progress_logger.info(message,
                         perc=perc, fqdn=fqdn, status=status, status_color=status_color,
                         worker=task_result.worker_name, task_time=elapsed_time)
    return task_result


# =====
def gen(args: cli_args.ShowGenOptions):
    """ Сгенерировать конфиг для устройств """
    connector = storage_connector.get()
    storage_opts = connector.opts().from_cli_opts(args)
    with connector.storage()(storage_opts) as storage:
        loader = ann_gen.Loader(storage, args)
        stdin = args.stdin(storage=storage, filter_acl=args.filter_acl, config=None)

    filterer = filtering.filterer_connector.get()
    pool = Parallel(ann_gen.worker, args, stdin, loader, filterer).tune_args(args)
    if args.show_hosts_progress:
        pool.add_callback(log_host_progress_cb)

    return pool.run(loader.device_ids, args.tolerate_fails, args.strict_exit_code)


# =====
def _diff_file(old_text: Optional[str], new_text: Optional[str], context=3):
    old_lines = old_text.splitlines() if old_text else []
    new_lines = new_text.splitlines() if new_text else []
    context = max(len(old_lines), len(new_lines)) if context is None else context
    return list(difflib.unified_diff(old_lines, new_lines, n=context, lineterm=""))


def _diff_files(old_files, new_files, context=3):
    ret = {}
    for (path, (new_text, reload_data)) in new_files.items():
        old_text = old_files.get(path)
        is_new = old_text is None
        diff_lines = _diff_file(old_text, new_text, context=context)
        ret[path] = (diff_lines, reload_data, is_new)
    return ret


def patch(args: cli_args.ShowPatchOptions):
    """ Сгенерировать патч для устройств """
    global live_configs  # pylint: disable=global-statement
    connector = storage_connector.get()
    storage_opts = connector.opts().from_cli_opts(args)
    with connector.storage()(storage_opts) as storage:
        loader = ann_gen.Loader(storage, args)
        if args.config == "running":
            fetcher = annet.deploy.fetcher_connector.get()
            live_configs = fetcher.fetch(loader.devices, processes=args.parallel)
        stdin = args.stdin(storage=storage, filter_acl=args.filter_acl, config=args.config)

    filterer = filtering.filterer_connector.get()
    pool = Parallel(_patch_worker, args, stdin, loader, filterer).tune_args(args)
    if args.show_hosts_progress:
        pool.add_callback(log_host_progress_cb)
    return pool.run(loader.device_ids, args.tolerate_fails, args.strict_exit_code)


def _patch_worker(device_id, args: cli_args.ShowPatchOptions, stdin, loader: ann_gen.Loader, filterer: filtering.Filterer):
    for res, _, patch_tree in res_diff_patch(device_id, args, stdin, loader, filterer):
        new_files = res.get_new_files(args.acl_safe)
        new_json_fragment_files = res.get_new_file_fragments(args.acl_safe)
        if new_files:
            for path, (cfg_text, _cmds) in new_files.items():
                label = res.device.hostname + os.sep + path
                yield label, cfg_text, False
        elif res.old_json_fragment_files or new_json_fragment_files:
            for path, (new_json_cfg, _cmds) in new_json_fragment_files.items():
                label = res.device.hostname + os.sep + path
                old_json_cfg = res.old_json_fragment_files[path]
                json_patch = jsontools.make_patch(old_json_cfg, new_json_cfg)
                yield (
                    label,
                    jsontools.format_json(json_patch),
                    False,
                )
        elif patch_tree:
            yield (
                "%s.patch" % res.device.hostname,
                _format_patch_blocks(patch_tree, res.device.hw, args.indent),
                False,
            )


# =====
def res_diff_patch(device_id, args: cli_args.ShowPatchOptions, stdin, loader: ann_gen.Loader, filterer: filtering.Filterer) -> Iterable[
    Tuple[OldNewResult, Dict, Dict]]:
    connector = storage_connector.get()
    storage_opts = connector.opts().from_cli_opts(args)
    with connector.storage()(storage_opts) as storage:
        for res in ann_gen.old_new(
            args,
            storage,
            config=args.config,
            loader=loader,
            filterer=filterer,
            stdin=stdin,
            device_ids=[device_id],
            no_new=args.clear,
            do_files_download=True,
        ):
            old = res.get_old(args.acl_safe)
            new = res.get_new(args.acl_safe)
            new_json_fragment_files = res.get_new_file_fragments(args.acl_safe)

            device = res.device
            acl_rules = res.get_acl_rules(args.acl_safe)
            if res.old_json_fragment_files or new_json_fragment_files:
                yield res, None, None
            elif old is not None:
                (diff_tree, patch_tree) = _diff_and_patch(device, old, new, acl_rules, res.filter_acl_rules,
                                                          args.add_comments)
                yield res, diff_tree, patch_tree


def diff(args: cli_args.DiffOptions, loader: ann_gen.Loader, filterer: filtering.Filterer) -> Mapping[Device, Union[Diff, PCDiff]]:
    ret = {}
    connector = storage_connector.get()
    storage_opts = connector.opts().from_cli_opts(args)
    with connector.storage()(storage_opts) as storage:
        for res in ann_gen.old_new(
            args,
            storage,
            config=args.config,
            loader=loader,
            no_new=args.clear,
            do_files_download=True,
            device_ids=loader.device_ids,
            filterer=filterer,
        ):
            old = res.get_old(args.acl_safe)
            new = res.get_new(args.acl_safe)
            device = res.device
            acl_rules = res.get_acl_rules(args.acl_safe)
            new_files = res.get_new_files(args.acl_safe)
            new_json_fragment_files = res.get_new_file_fragments()
            pc_diff_files = []
            if res.old_files or new_files:
                pc_diff_files.extend(_pc_diff(device.hostname, res.old_files, new_files))
            if res.old_json_fragment_files or new_json_fragment_files:
                pc_diff_files.extend(_json_fragment_diff(device.hostname, res.old_json_fragment_files, new_json_fragment_files))

            if pc_diff_files:
                pc_diff_files.sort(key=lambda f: f.label)
                ret[device] = PCDiff(hostname=device.hostname, diff_files=pc_diff_files)
            elif old is not None:
                rb = rulebook.get_rulebook(device.hw)
                diff_tree = patching.make_diff(old, new, rb, [acl_rules, res.filter_acl_rules])
                diff_tree = patching.strip_unchanged(diff_tree)
                ret[device] = diff_tree
    return ret


def collapse_texts(texts: Mapping[str, str]) -> Mapping[Tuple[str, ...], str]:
    """
    Группировка текстов.
    :param texts:
    :return: словарь с несколькими хостнеймами в ключе.
    """
    diffs_with_orig = {key: [value, value.splitlines()] for key, value in texts.items()}
    res = {}
    for _, collapsed_diff_iter in groupby(sorted(diffs_with_orig.items(), key=lambda x: (x[0], x[1][1])),
                                          key=lambda x: x[1][1]):
        collapsed_diff = list(collapsed_diff_iter)
        res[tuple(x[0] for x in collapsed_diff)] = collapsed_diff[0][1][0]

    return res


class DeployerJob(abc.ABC):
    def __init__(self, device, args: cli_args.DeployOptions):
        self.args = args
        self.device = device
        self.add_comments = False
        self.diff_lines = []
        self.cmd_lines: List[str] = []
        self.deploy_cmds = odict()
        self.diffs = {}
        self.failed_configs = {}
        self._has_diff = False

    @abc.abstractmethod
    def parse_result(self, res):
        pass

    def collapseable_diffs(self):
        return {}

    def has_diff(self):
        return self._has_diff

    @staticmethod
    def from_device(device, args: cli_args.DeployOptions):
        if device.hw.vendor == "pc":
            return PCDeployerJob(device, args)
        return CliDeployerJob(device, args)


class CliDeployerJob(DeployerJob):
    def parse_result(self, res: OldNewResult):
        device = res.device
        old = res.get_old(self.args.acl_safe)
        new = res.get_new(self.args.acl_safe)
        acl_rules = res.get_acl_rules(self.args.acl_safe)
        err = res.err

        if err:
            self.failed_configs[device.fqdn] = err
            return

        (diff_obj, patch_tree) = _diff_and_patch(device, old, new, acl_rules,
                                                 res.filter_acl_rules, self.add_comments,
                                                 do_commit=not self.args.dont_commit)
        cmds = tabparser.make_formatter(device.hw, indent="").cmd_paths(patch_tree)
        if not cmds:
            return
        self._has_diff = True
        self.diffs[device] = diff_obj
        self.cmd_lines.extend(["= %s " % device.hostname, ""])
        self.cmd_lines.extend(map(itemgetter(-1), cmds))
        self.cmd_lines.append("")
        deployer_driver = annet.deploy.driver_connector.get()
        self.deploy_cmds[device] = deployer_driver.apply_deploy_rulebook(
            device.hw, cmds,
            do_commit=not self.args.dont_commit
        )
        for cmd in deployer_driver.build_exit_cmdlist(device.hw):
            self.deploy_cmds[device].add_cmd(cmd)

    def collapseable_diffs(self):
        return self.diffs


class PCDeployerJob(DeployerJob):
    def parse_result(self, res: ann_gen.OldNewResult):
        device = res.device
        old_files = res.old_files
        new_files = res.get_new_files(self.args.acl_safe)
        old_json_fragment_files = res.old_json_fragment_files
        new_json_fragment_files = res.get_new_file_fragments(self.args.acl_safe)
        err = res.err

        if err:
            self.failed_configs[device.fqdn] = err
            return
        elif not new_files and not new_json_fragment_files:
            return

        upload_files: Dict[str, bytes] = {}
        reload_cmds: Dict[str, bytes] = {}
        generator_types: Dict[str, GeneratorType] = {}
        for generator_type, pc_files in [(GeneratorType.ENTIRE, new_files), (GeneratorType.JSON_FRAGMENT, new_json_fragment_files)]:
            for file, (file_content_or_json_cfg, cmds) in pc_files.items():
                if generator_type == GeneratorType.ENTIRE:
                    file_content: str = file_content_or_json_cfg
                    diff_content = "\n".join(_diff_file(old_files.get(file), file_content))
                else:  # generator_type == GeneratorType.JSON_FRAGMENT
                    old_json_cfg = old_json_fragment_files[file]
                    json_patch = jsontools.make_patch(old_json_cfg, file_content_or_json_cfg)
                    file_content = jsontools.format_json(json_patch)
                    old_text = jsontools.format_json(old_json_cfg)
                    new_text = jsontools.format_json(file_content_or_json_cfg)
                    diff_content = "\n".join(_diff_file(old_text, new_text))
                if diff_content:
                    self._has_diff = True
                upload_files[file], reload_cmds[file] = file_content.encode(), cmds.encode()
                generator_types[file] = generator_type
                self.cmd_lines.append("= Deploy cmds %s/%s " % (device.hostname, file))
                self.cmd_lines.extend([cmds, ""])
                self.cmd_lines.append("= %s/%s " % (device.hostname, file))
                self.cmd_lines.extend([file_content, ""])
                self.diff_lines.append("= %s/%s " % (device.hostname, file))
                self.diff_lines.extend([diff_content, ""])

        if upload_files:
            self.deploy_cmds[device] = {
                "files": upload_files,
                "cmds": reload_cmds,
                "generator_types": generator_types,
            }
            self.diffs[device] = upload_files
            deployer_driver = annet.deploy.driver_connector.get()
            before, after = deployer_driver.build_configuration_cmdlist(device.hw)
            for cmd in deployer_driver.build_exit_cmdlist(device.hw):
                after.add_cmd(cmd)
            cmds_pre_files = {}
            for file in self.deploy_cmds[device]["files"]:
                if before:
                    cmds_pre_files[file] = "\n".join(map(str, before)).encode(encoding="utf-8")
                self.deploy_cmds[device]["cmds"][file] += "\n".join(map(str, after)).encode(encoding="utf-8")
            self.deploy_cmds[device]["cmds_pre_files"] = cmds_pre_files


class Deployer:
    def __init__(self, args: cli_args.DeployOptions):
        self.args = args

        self.cmd_lines = []
        self.deploy_cmds = odict()
        self.diffs = {}
        self.failed_configs: Dict[str, Exception] = {}
        self.fqdn_to_device: Dict[str, Device] = {}
        self.empty_diff_hostnames: Set[str] = set()

        self._collapseable_diffs = {}
        self._diff_lines: List[str] = []
        self._filterer = filtering.filterer_connector.get()

    def parse_result(self, job: DeployerJob, result: ann_gen.OldNewResult):
        entire_reload = self.args.entire_reload
        logger = get_logger(job.device.hostname)

        job.parse_result(result)
        self.failed_configs.update(job.failed_configs)

        if job.has_diff() or entire_reload is entire_reload.force:
            self.cmd_lines.extend(job.cmd_lines)
            self.deploy_cmds.update(job.deploy_cmds)
            self.diffs.update(job.diffs)

            self.fqdn_to_device[result.device.fqdn] = result.device
            self._collapseable_diffs.update(job.collapseable_diffs())
            self._diff_lines.extend(job.diff_lines)
        else:
            logger.info("empty diff")

    def diff_lines(self) -> List[str]:
        diff_lines = []
        diff_lines.extend(self._diff_lines)
        for devices, diff_obj in ann_diff.collapse_diffs(self._collapseable_diffs).items():
            if not diff_obj:
                self.empty_diff_hostnames.update(dev.hostname for dev in devices)
            if not self.args.no_ask_deploy:
                # разобъем список устройств на несколько линий
                dest_name = ""
                try:
                    _, term_columns_str = os.popen("stty size", "r").read().split()
                    term_columns = int(term_columns_str)
                except Exception:
                    term_columns = 2 ** 32
                fqdns = [dev.hostname for dev in devices]
                while fqdns:
                    fqdn = fqdns.pop()
                    if len(dest_name) == 0:
                        dest_name = "= %s" % fqdn
                    elif len(dest_name) + len(fqdn) < term_columns:
                        dest_name = "%s, %s" % (dest_name, fqdn)
                    else:
                        diff_lines.extend([dest_name])
                        dest_name = "= %s" % fqdn
                    if not fqdns:
                        diff_lines.extend([dest_name, ""])
            else:
                dest_name = "= %s" % ", ".join([dev.hostname for dev in devices])
                diff_lines.extend([dest_name, ""])

            for line in tabparser.make_formatter(devices[0].hw).diff(diff_obj):
                diff_lines.append(line)
            diff_lines.append("")
        return diff_lines

    def ask_deploy(self) -> str:
        return self._ask("y", annet.deploy.AskConfirm(
            text="\n".join(self.diff_lines()),
            alternative_text="\n".join(self.cmd_lines),
        ))

    def ask_rollback(self) -> str:
        return self._ask("n", annet.deploy.AskConfirm(
            text="Execute rollback?\n",
            alternative_text="",
        ))

    def _ask(self, default_ans: str, ask: annet.deploy.AskConfirm) -> str:
        # если filter_acl из stdin то с ним уже не получится работать как с терминалом
        ans = default_ans
        if not self.args.no_ask_deploy:
            try:
                if not os.isatty(sys.stdin.fileno()):
                    pts_path = os.ttyname(sys.stdout.fileno())
                    pts = open(pts_path, "r")  # pylint: disable=consider-using-with
                    os.dup2(pts.fileno(), sys.stdin.fileno())
            except OSError:
                pass
            ans = ask.loop()
        return ans

    def check_diff(self, result: annet.deploy.DeployResult, storage: Storage):
        global live_configs  # pylint: disable=global-statement
        success_hosts = [
            host.split(".", 1)[0] for (host, hres) in result.results.items()
            if (not isinstance(hres, Exception) and
                host not in self.empty_diff_hostnames and
                not self.fqdn_to_device[host].is_pc())
        ]
        diff_args = self.args.copy_from(
            self.args,
            config="running",
            query=success_hosts,
        )
        if diff_args.query:
            live_configs = None
            loader = ann_gen.Loader(storage, diff_args, no_empty_warning=True)

            diffs = diff(diff_args, loader, self._filterer)
            non_pc_diffs = {dev: diff for dev, diff in diffs.items() if not isinstance(diff, PCDiff)}
            devices_to_diff = ann_diff.collapse_diffs(non_pc_diffs)
            devices_to_diff.update({(dev,): diff for dev, diff in diffs.items() if isinstance(diff, PCDiff)})
        else:
            devices_to_diff = {}
        for devices, diff_obj in devices_to_diff.items():
            if diff_obj:
                for dev in devices:
                    self.failed_configs[dev.fqdn] = Warning("Deploy OK, but diff still exists")
                if isinstance(diff_obj, PCDiff):
                    for diff_file in diff_obj.diff_files:
                        print_err_label(diff_file.label)
                        print("\n".join(format_file_diff(diff_file.diff_lines)))
                else:
                    output_driver = output_driver_connector.get()
                    dest_name = ", ".join([output_driver.cfg_file_names(dev)[0] for dev in devices])
                    print_err_label(dest_name)
                    _print_pre_as_diff(patching.make_pre(diff_obj), diff_args.show_rules, diff_args.indent)


def deploy(args: cli_args.DeployOptions) -> ExitCode:
    """ Сгенерировать конфиг для устройств и задеплоить его """
    ret: ExitCode = 0
    deployer = Deployer(args)
    connector = storage_connector.get()
    storage_opts = connector.opts().from_cli_opts(args)
    with connector.storage()(storage_opts) as storage:
        global live_configs  # pylint: disable=global-statement
        loader = ann_gen.Loader(storage, args)
        filterer = filtering.filterer_connector.get()
        fetcher = annet.deploy.fetcher_connector.get()
        deploy_driver = annet.deploy.driver_connector.get()
        live_configs = fetcher.fetch(devices=loader.devices, processes=args.parallel)
        pool = ann_gen.OldNewParallel(storage, args, loader, filterer)

        for res in pool.generated_configs(loader.device_ids):
            # Меняем exit code если хоть один device ловил exception
            if res.err is not None:
                ret |= 2 ** 3
            job = DeployerJob.from_device(res.device, args)
            deployer.parse_result(job, res)

        deploy_cmds = deployer.deploy_cmds
        result = annet.deploy.DeployResult(hostnames=[], results={}, durations={}, original_states={})
        if deploy_cmds:
            ans = deployer.ask_deploy()
            if ans != "y":
                return 2 ** 2
            result = annet.lib.do_async(deploy_driver.bulk_deploy(deploy_cmds, args))

        rolled_back = False
        rollback_cmds = {deployer.fqdn_to_device[x]: cc for x, cc in result.original_states.items() if cc}
        if args.rollback and rollback_cmds:
            ans = deployer.ask_rollback()
            if rollback_cmds and ans == "y":
                rolled_back = True
                annet.lib.do_async(deploy_driver.bulk_deploy(rollback_cmds, args))

        if not args.no_check_diff and not rolled_back:
            deployer.check_diff(result, storage)

        if deployer.failed_configs:
            result.add_results(deployer.failed_configs)
            ret |= 2 ** 1

        annet.deploy.show_bulk_report(result.hostnames, result.results, result.durations, log_dir=None)
        for host_result in result.results.values():
            if isinstance(host_result, Exception):
                ret |= 2 ** 0
                break
        return ret


def file_diff(args: cli_args.FileDiffOptions):
    """ Создать дифф по рулбуку между файлами или каталогами """
    old_new = list(_read_old_new_cfgdumps(args))
    pool = Parallel(file_diff_worker, args).tune_args(args)
    return pool.run(old_new, tolerate_fails=True)


def file_diff_worker(old_new: Tuple[str, str], args: cli_args.FileDiffOptions) -> Generator[
    Tuple[str, str, bool], None, None]:
    old_path, new_path = old_new
    if os.path.isdir(old_path) and os.path.isdir(new_path):
        hostname = os.path.basename(new_path)
        new_files = {relative_cfg_path: (cfg_text, "") for relative_cfg_path, cfg_text in
                     ann_gen.load_pc_config(new_path).items()}
        old_files = ann_gen.load_pc_config(old_path)
        for diff_file in _pc_diff(hostname, old_files, new_files):
            diff_text = (
                "\n".join(diff_file.diff_lines)
                if args.no_color
                else "\n".join(format_file_diff(diff_file.diff_lines))
            )
            if diff_text:
                yield diff_file.label, diff_text, False
    else:
        dest_name, old, new, hw = _read_old_new_hw(old_path, new_path, args)
        _, __, pre, ___ = _read_old_new_diff_patch(old, new, hw, add_comments=False)
        diff_lines = ann_diff.gen_pre_as_diff(pre, args.show_rules, args.indent, args.no_color)
        diff_text = "".join(diff_lines)
        if diff_text:
            yield dest_name, diff_text, False


@tracing.function
def file_patch(args: cli_args.FilePatchOptions):
    """ Создать патч между файлами или каталогами """
    old_new = list(_read_old_new_cfgdumps(args))
    pool = Parallel(file_patch_worker, args).tune_args(args)
    return pool.run(old_new, tolerate_fails=True)


def file_patch_worker(old_new: Tuple[str, str], args: cli_args.FileDiffOptions) -> Generator[
    Tuple[str, str, bool], None, None]:
    old_path, new_path = old_new
    if os.path.isdir(old_path) and os.path.isdir(new_path):
        for relative_cfg_path, cfg_text in ann_gen.load_pc_config(new_path).items():
            label = os.path.join(os.path.basename(new_path), relative_cfg_path)
            yield label, cfg_text, False
    else:
        dest_name, old, new, hw = _read_old_new_hw(old_path, new_path, args)
        _, __, ___, patch_tree = _read_old_new_diff_patch(old, new, hw, args.add_comments)
        patch_text = _format_patch_blocks(patch_tree, hw, args.indent)
        if patch_text:
            yield dest_name, patch_text, False


def _pc_diff(hostname: str, old_files: Dict[str, str], new_files: Dict[str, str]) -> Generator[PCDiffFile, None, None]:
    sorted_lines = sorted(_diff_files(old_files, new_files).items())
    for (path, (diff_lines, _reload_data, is_new)) in sorted_lines:
        if not diff_lines:
            continue
        label = hostname + os.sep + path
        if is_new:
            label = LABEL_NEW_PREFIX + label
        yield PCDiffFile(label=label, diff_lines=diff_lines)


def _json_fragment_diff(
        hostname: str,
        old_files: Dict[str, Any],
        new_files: Dict[str, Tuple[Any, Optional[str]]],
) -> Generator[PCDiffFile, None, None]:
    def jsonify_multi(files):
        return {
            path: jsontools.format_json(cfg)
            for path, cfg in files.items()
        }

    def jsonify_multi_with_cmd(files):
        ret = {}
        for path, cfg_reload_cmd in files.items():
            cfg, reload_cmd = cfg_reload_cmd
            ret[path] = (jsontools.format_json(cfg), reload_cmd)
        return ret
    jold, jnew = jsonify_multi(old_files), jsonify_multi_with_cmd(new_files)
    return _pc_diff(hostname, jold, jnew)


def guess_hw(config_text: str):
    """Пытаемся угадать вендора и hw на основе
    текста конфига и annet/rulebook/texts/*.rul"""
    scores = {}
    hw_provider = hardware_connector.get()
    for vendor in VENDOR_REVERSES:
        hw = hw_provider.vendor_to_hw(vendor)
        fmtr = tabparser.make_formatter(hw)
        rb = rulebook.get_rulebook(hw)
        config = tabparser.parse_to_tree(config_text, fmtr.split)
        pre = patching.make_pre(patching.make_diff({}, config, rb, []))
        metric = _count_pre_score(pre)
        scores[metric] = hw
    max_score = max(scores.keys())
    hw = scores[max_score]
    return hw, max_score


def _count_pre_score(top_pre) -> float:
    """Обходим вширь pre-конфиг
    и подсчитываем количество заматчившихся
    правил на каждом из уровней.

    Чем больше результирующий приоритет
    тем больше рулбук соответсвует конфигу.
    """
    score = 0
    scores = []
    cur, child = [top_pre], []
    while cur:
        for pre in cur.pop().values():
            score += 1
            for item in pre["items"].values():
                for op in [Op.ADDED, Op.AFFECTED, Op.REMOVED]:
                    child += [x["children"] for x in item[op]]
        if not cur:
            scores.append(score)
            score = 0
            cur, child = child, []
    result = 0
    for i in reversed(scores):
        result <<= i.bit_length()
        result += i
    if result > 0:
        result = 1 - (1 / result)
    return float(result)



from annet.annlib.patching import (  # pylint: disable=unused-import
    AclError,
    AclNotExclusiveError,
)
from annet.annlib.patching import Orderer as BaseOrderer
from annet.annlib.patching import (  # pylint: disable=unused-import
    PatchTree,
    apply_acl,
    apply_diff_rb,
    make_diff,
    make_patch,
    make_pre,
    strip_unchanged,
)

from annet import rulebook


class Orderer(BaseOrderer):
    @classmethod
    def from_hw(cls, hw):
        return cls(
            rulebook.get_rulebook(hw)["ordering"],
            hw.vendor,
        )



import asyncio
import dataclasses
import enum
import faulthandler
import inspect
import multiprocessing as mp
import os
import pickle
import platform
import queue
import re
import signal
import sys
import tempfile
import time
import traceback
import warnings
from typing import Any, List, Optional, Type
from uuid import uuid4

from contextlog import get_logger

import annet
from annet import tracing
from annet.lib import catch_ctrl_c, find_exc_in_stack
from annet.output import capture_output
from annet.tracing import tracing_connector


class PoolWorkerTaskType(enum.Enum):
    INVOKE = "invoke"
    STOP = "stop"


@dataclasses.dataclass(frozen=True)
class PoolWorkerTask:
    type: PoolWorkerTaskType
    payload: Optional[Any] = None


class PickleSafeException(Exception):
    """An exception that can be safely pickled and passed between processes."""

    def __init__(self, orig_exc_cls: Type[Exception], orig_exc_msg: str, device_id: str, formatted_output: str) -> None:
        self.orig_exc_cls = orig_exc_cls
        self.orig_exc_msg = orig_exc_msg
        self.device_id = device_id
        self.formatted_output = formatted_output
        super().__init__(orig_exc_cls, orig_exc_msg, device_id, formatted_output)

    def __repr__(self) -> str:
        orig_cls_name = self.orig_exc_cls.__name__
        orig_msg_repr = repr(self.orig_exc_msg)
        return f"PickleSafeException<{orig_cls_name}({orig_msg_repr})>"

    def __str__(self) -> str:
        pretty_lines: List[str] = []
        pretty_lines.append(f"An exception for device {self.device_id}:")
        pretty_lines.append(self.formatted_output)
        return "\n".join(pretty_lines)

    @staticmethod
    def from_exc(orig_exc: Exception, device_id: str, formatted_output: str) -> "PickleSafeException":
        return PickleSafeException(orig_exc.__class__, str(orig_exc), device_id, formatted_output)


@catch_ctrl_c
def pool_worker(pool, index, task_queue, done_queue, context_carrier):
    faulthandler.register(signal.SIGUSR1)

    tracing_connector.get().attach_context(tracing_connector.get().extract_context(context_carrier))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=DeprecationWarning)
        try:
            asyncio.get_event_loop().close()
        except Exception:
            pass
        asyncio.set_event_loop(asyncio.new_event_loop())

    return _pool_worker(pool, index, task_queue, done_queue)


@tracing.function(flush=True)
def _pool_worker(pool, index, task_queue, done_queue):
    worker_id = uuid4().hex

    pool_span = tracing_connector.get().get_current_span()
    if pool_span:
        pool_span.update_name("pool_worker")
        pool_span.set_attribute("worker.id", worker_id)
        pool_span.set_attribute("worker.index", index)

    cap_stdout = tempfile.TemporaryFile(mode="w+") if pool.capture_output else None
    cap_stderr = tempfile.TemporaryFile(mode="w+") if pool.capture_output else None

    _logger = get_logger()
    tasks_done = 0
    device_ids = []
    worker_name = mp.current_process().name

    while True:
        task: PoolWorkerTask = task_queue.get()
        if task.type == PoolWorkerTaskType.STOP:
            _logger.debug("I received STOP, terminating...")
            return

        device_id: Optional[str] = None
        if isinstance(task.payload, tuple):
            if len(task.payload) > 0:
                match = re.search(r"([^/]+).cfg", task.payload[0])
                if match:
                    device_id = match.group(1)
        else:
            device_id = str(task.payload)

        if device_id:
            device_ids.append(device_id)

        if pool_span:
            pool_span.set_attribute("device.ids", device_ids)

        task_result = TaskResult(worker_name, task.payload)
        task_result.extra["start_time"] = time.monotonic()
        ret_exc = None

        try:
            with tracing_connector.get().start_as_current_span(
                "pool_worker.invoke",
            ) as span:
                if device_id:
                    span.set_attribute("device.id", device_id)

                name = "invoke"
                invoke_span_ctx = tracing_connector.get().start_as_linked_span(name, tracer_name=__name__)
                capture_output_ctx = capture_output(cap_stdout, cap_stderr)

                with invoke_span_ctx as invoke_span, capture_output_ctx as _:
                    invoke_trace_id = invoke_span.get_span_context().trace_id
                    if invoke_trace_id:
                        span.set_attribute(
                            "link",
                            f"https://t.yandex-team.ru/trace/{invoke_trace_id:x}"
                        )
                    invoke_span.set_attribute("func", pool.func.__name__)
                    invoke_span.set_attribute("worker.id", worker_id)
                    if device_id:
                        invoke_span.set_attribute("device.id", device_id)

                    _logger.warning("Worker-%d start invoke %s", index, device_id)
                    task_result.result = invoke_retry(
                        pool.func,
                        pool.net_retry,
                        task.payload,
                        *pool.args,
                        **pool.kwargs
                    )
                    _logger.warning("Worker-%d finish invoke %s", index, device_id)

                    # Check if the result is picklable and throw an exception if not.
                    # Otherwise the exception will be thrown inside the multiprocessing
                    # code and we won't be able to handle it.
                    pickle.dumps(task_result.result)
        except KeyboardInterrupt:  # pylint: disable=try-except-raise
            raise
        except Exception as exc:
            safe_exc = PickleSafeException.from_exc(exc, task.payload, traceback.format_exc())
            ret_exc = safe_exc
            task_result.exc = safe_exc
            task_result.result = None
        if pool.capture_output:
            task_result.extra["cap_stdout"] = cap_stdout.read()
            task_result.extra["cap_stderr"] = cap_stderr.read()

        results = list(pool._run_callbacks(task_result, in_thread=True))  # pylint: disable=protected-access
        done_queue.put((worker_name, task, results, ret_exc))

        tasks_done += 1
        if pool.max_tasks and tasks_done >= pool.max_tasks:
            _logger.debug("Maximum tasks limit reached. Now I can retire")
            tracing_connector.get().force_flush()
            sys.exit(9)


class TaskResult:
    def __init__(self, worker_name, device_id, result=None, exc=None):
        self.worker_name = worker_name
        self.device_id = device_id
        self.result = result
        self.exc = exc
        self.extra = {}

    def __repr__(self):
        return "TaskResult(worker_name=%s, device_id=%s, result=%s, exc=%s, extra=%s)" % (
            self.worker_name, self.device_id, self.result, self.exc, self.extra)


class Parallel:
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.callbacks = []
        self.in_thread_callbacks = []
        self.parallel = mp.cpu_count()
        self.task_timeout = 1800  # maximum seconds to wait for next task done
        self.max_tasks = 25
        self.tasks_done = 0
        self.net_retry = 3
        self.capture_output = False

    def tune(self, **kwargs):
        for (kw, arg) in kwargs.items():
            if not hasattr(self, kw):
                raise ValueError("Can not tune Parallel.%s: attribute doesn't exist" % kw)
            setattr(self, kw, arg)
        return self

    def tune_args(self, args):
        for kw in args._enum_args():  # pylint: disable=protected-access
            if hasattr(self, kw):
                setattr(self, kw, getattr(args, kw))
        return self

    # func prototype: func(parallel_object, worker_name, ret) -> ret(s)
    def add_callback(self, func, in_thread=False):
        if not callable(func):
            raise annet.ExecError("callback must be a callable object or function")
        if in_thread:
            self.in_thread_callbacks.append(func)
        else:
            self.callbacks.append(func)
        return self

    def _run_callbacks(self, task_result: TaskResult, in_thread=False):
        task_results = [task_result]
        cbs = self.in_thread_callbacks if in_thread else self.callbacks
        for cb in cbs:
            task_results = [cb_result for result in task_results for cb_result in self._cb_wrapper(cb, result)]
        yield from task_results

    def _cb_wrapper(self, cb, task_result: TaskResult):
        task_results = []
        try:
            if inspect.isgeneratorfunction(cb):
                future = cb(self, task_result)
                while True:
                    cb_result = next(future)
                    if isinstance(cb_result, TaskResult):
                        task_results.append(cb_result)
            else:
                cb_result = cb(self, task_result)
                if isinstance(cb_result, TaskResult):
                    task_results = [cb_result]
        except StopIteration:
            pass
        except Exception as exc:
            exc.device_id = task_result.device_id
            exc.formatted_output = traceback.format_exc()
            task_result.exc = exc
            task_results = [task_result]
        yield from task_results

    @tracing.function
    def run(self, device_ids, tolerate_fails=True, strict_error_code=False):
        success, fail = {}, {}
        for task_result in self.irun(device_ids, tolerate_fails):
            if task_result.exc is not None:
                fail[task_result.device_id] = task_result.exc
            else:
                success[task_result.device_id] = task_result.result
        if strict_error_code and fail:
            raise RuntimeError("failed for %d/%d devices" % (len(fail), len(device_ids)))
        return success, fail

    def irun(self, device_ids, tolerate_fails=True):
        _logger = get_logger()
        self.tasks_done = 0
        pool_size = self.parallel if len(device_ids) > self.parallel else len(device_ids)

        span = tracing_connector.get().get_current_span()
        if span:
            span.set_attribute("pool_size", pool_size)

        # single process way
        if pool_size == 1:
            cap_stdout = tempfile.TemporaryFile(mode="w+") if self.capture_output else None
            cap_stderr = tempfile.TemporaryFile(mode="w+") if self.capture_output else None
            worker_name = mp.current_process().name

            for device_id in device_ids:
                task_result = TaskResult(worker_name, device_id)
                task_result.extra["start_time"] = time.monotonic()
                try:
                    with capture_output(cap_stdout, cap_stderr):
                        task_result.result = invoke_retry(
                            self.func,
                            self.net_retry,
                            device_id,
                            *self.args,
                            **self.kwargs
                        )
                except Exception as exc:
                    if not tolerate_fails:
                        raise
                    exc.device_id = device_id
                    exc.formatted_output = traceback.format_exc()
                    task_result.exc = exc
                if self.capture_output:
                    task_result.extra["cap_stdout"] = cap_stdout.read()
                    task_result.extra["cap_stderr"] = cap_stderr.read()
                self.tasks_done += 1

                yield from [
                    result
                    for in_thread_result in self._run_callbacks(task_result, in_thread=True)
                    for result in self._run_callbacks(in_thread_result)
                ]
        else:
            # multiple processes way
            _logger.info("creating process pool with %d workers", pool_size)
            task_queue = mp.Queue()
            done_queue = mp.Queue()
            for device_id in device_ids:
                task_queue.put(PoolWorkerTask(type=PoolWorkerTaskType.INVOKE, payload=device_id))

            context_carrier = {}
            tracing_connector.get().inject_context(context_carrier)

            pool = {}
            for index in range(pool_size):
                task_queue.put(PoolWorkerTask(type=PoolWorkerTaskType.STOP))
                worker_name = "Worker-%d" % index
                worker_args = (self, index, task_queue, done_queue, context_carrier)

                worker = mp.Process(
                    name=worker_name,
                    target=pool_worker,
                    args=worker_args
                )
                pool[worker_name] = worker
                worker.start()
                _logger.debug("Worker '%s' has been created with PID %d", worker_name, worker.pid)

            last_task_ts = time.monotonic()
            while True:
                exc = None
                worker_name = None
                in_thread_results = None

                queue_empty = False
                try:
                    worker_name, _, in_thread_results, exc = done_queue.get(True, 1)
                    last_task_ts = time.monotonic()
                except queue.Empty:
                    queue_empty = True

                retired_workers, failed_workers = self._check_children(pool)

                terminate_by_timeout = False
                terminate_exc = None
                if queue_empty:
                    if time.monotonic() - last_task_ts > self.task_timeout:
                        # timeout hit
                        terminate_by_timeout = True
                        terminate_exc = annet.ExecError()

                if not tolerate_fails:
                    if exc is not None:
                        # worker returned exception
                        terminate_exc = exc
                    elif failed_workers:
                        # some workers exited with non-zero (and non-9) code
                        terminate_exc = annet.ExecError(f"Workers {failed_workers} exited with error")

                if terminate_exc is not None:
                    for (name, worker) in pool.items():
                        if worker.exitcode is None:
                            if terminate_by_timeout:
                                os.kill(worker.pid, signal.SIGUSR1)  # force dump stacktrace
                                time.sleep(10)
                            worker.terminate()
                            _logger.warning("Worker '%s' (PID: %d) has been terminated", name, worker.pid)
                        worker.join()
                    raise terminate_exc

                if not queue_empty:
                    self.tasks_done += 1

                    qsize = "unknown"
                    if platform.system() != "Darwin":
                        # Not implemented for macOS
                        qsize = str(done_queue.qsize())

                    _logger.debug("Got a result from worker '%s', qsize is %s", worker_name, qsize)

                    yield from [
                        result
                        for in_thread_result in in_thread_results
                        for result in self._run_callbacks(in_thread_result)
                    ]

                if not pool:
                    break

                for name in retired_workers:
                    _logger.debug("Worker '%s' has retired. Restart it", name)
                    pool[name] = mp.Process(name=name, target=pool_worker, args=worker_args)
                    pool[name].start()
            task_queue.close()
            done_queue.close()

    def _check_children(self, pool):
        _logger = get_logger()
        retired_workers = []
        failed_workers = []
        for name in list(pool.keys()):
            exitcode = pool[name].exitcode
            if exitcode is None:
                continue
            if exitcode == 9:
                retired_workers.append(name)
            elif exitcode != 0:
                _logger.error("Worker '%s' (PID: %d) has exited with non-zero exit code %d", name, pool[name].pid,
                              exitcode)
                failed_workers.append(name)
            else:
                _logger.debug("Worker '%s' (PID: %d) has been reaped with exitcode %d", name, pool[name].pid, exitcode)
            pool[name].join()
            if exitcode != 9:
                del pool[name]
        return retired_workers, failed_workers


def invoke_retry(func, net_retry, *args, **kwargs):
    attempt = 0
    while True:
        try:
            result = func(*args, **kwargs)
            if inspect.isgenerator(result):
                result = list(result)
            return result
        except Exception as exc:
            if not find_exc_in_stack(exc, (BrokenPipeError, ConnectionResetError)):
                raise
            if attempt >= net_retry:
                raise
            attempt += 1



import collections
import re
import typing

from . import patching, tabparser
from .diff import diff_ops, ops_sign
from .rbparser import acl

UnifiedInputConfig = str  # Конфиг классических сетевых устройств
FileInputConfig = typing.Dict[str, typing.Any]  # Конфиг вайтбоксов и серверов
InputConfig = typing.Union[UnifiedInputConfig, FileInputConfig]

Acl = typing.Dict[str, typing.Any]

UnifiedConfigTree = typing.OrderedDict[str, typing.Any]
FileConfigTree = typing.Dict[str, typing.Any]
ConfigTree = typing.Union[UnifiedConfigTree, FileConfigTree]

DiffTree = typing.OrderedDict[str, typing.Any]
UnifiedDiff = typing.List[typing.Tuple[str, str, typing.List, typing.Optional[int]]]
FileConfigDiff = typing.Dict[str, typing.Any]
Diff = typing.Union[UnifiedDiff, FileConfigDiff]


def make_acl(text: str, vendor: str) -> Acl:
    return acl.compile_acl_text(text, vendor)


def filter_config(acl: Acl, fmtr: tabparser.CommonFormatter, input_config: InputConfig) -> InputConfig:
    if isinstance(input_config, str):
        config: ConfigTree = tabparser.parse_to_tree(input_config, fmtr.split)
        config = patching.apply_acl(config, acl, fatal_acl=False)
        config = fmtr.join(config)
    else:
        config = typing.cast(input_config, FileConfigTree)
        config = apply_acl_fileconfig(input_config, acl)
    return config


def filter_diff(acl: Acl, fmtr: tabparser.CommonFormatter, input_config: InputConfig) -> InputConfig:
    if isinstance(input_config, str):
        input_config = shift_op(input_config)
        diff_tee: DiffTree = tabparser.parse_to_tree(input_config, fmtr.split)
        diff: Diff = tree_to_diff(diff_tee)
        diff = patching.apply_acl_diff(diff, acl)
        config = fmtr.join(diff_to_tree(diff))
        config = unshift_op(config)
        config = config.rstrip()
    else:
        config = typing.cast(input_config, FileConfigTree)
        config = apply_acl_fileconfig(input_config, acl)
    return config


def filter_patch(acl: Acl, fmtr: tabparser.CommonFormatter, text: str) -> str:
    return filter_config(acl, fmtr, text)


# NOCDEV-6378 на патч для Juniper/Nokia нельзя просто так наложить filter_acl
def filter_patch_jun_nokia(diff_filtered: InputConfig, fmtr: tabparser.CommonFormatter, text: str) -> str:
    """
    Накладываем ACL на патчи для Juniper/Nokia

    Поскольку в патче уже потерена иерархия команд - они развернуты в строки типа
    Нужна дополнительная информация о изначальном конфиге, которую можно подсмотреть в дифе
        set interface et-0/0/0 unit ....
        delete interface et-0/0/0 unit ....
        /configure port 1/1/c17/1 ...
        /configure delete port 1/1/c18/1 ...
    """
    diff_tree_stripped: DiffTree = tabparser.parse_to_tree(strip_op(diff_filtered), fmtr.split)
    _tree_expand_lists_nokia_jun(diff_tree_stripped)
    patch_lines_passed = []
    for patch_line in text.split("\n"):
        patch_parts = [x for x in patch_line.split(" ") if x]
        diff_current = diff_tree_stripped
        # strip set|delete|/configure
        while patch_parts and patch_parts[0] in {"/configure", "set", "delete"}:
            patch_parts = patch_parts[1:]
        while patch_parts and diff_current:
            for i in range(len(patch_parts), -1, -1):
                key = " ".join(patch_parts[:i])
                # consume parts and go down in diff hierarchy
                if key in diff_current:
                    patch_parts = patch_parts[i:]
                    diff_current = diff_current[key]
                    break
            # no progress has been made
            else:
                break
        if not patch_parts:
            patch_lines_passed.append(patch_line)
    return "\n".join(patch_lines_passed)


def apply_acl_fileconfig(config, rules):
    passed = {}
    for (filename, filecontent) in config.items():
        (match, _) = patching.match_row_to_acl(filename, rules)
        if match:
            if not (match["is_reverse"] and match["attrs"]["cant_delete"]):
                passed[filename] = filecontent
    return passed


def get_op(line: str) -> typing.Tuple[str, str, str]:
    op = " "
    indent = ""
    opidx = -1
    rowstart = 0
    for rowstart in range(len(line)):
        if line[rowstart] not in diff_ops:
            break
    for opidx in range(rowstart):
        if line[opidx] != " ":
            break
    if opidx >= 0:
        op = line[opidx]
        indent = line[:opidx] + line[opidx + 1:rowstart]
    if op != " ":
        indent = indent + " "
    return op, indent, line[rowstart:]


def shift_op(text: str) -> str:
    ret = ""
    for line in text.split("\n"):
        op, indent, line = get_op(line)
        ret += indent + op + line + "\n"
    return ret


def unshift_op(text: str) -> str:
    ret = ""
    for line in text.split("\n"):
        op, indent, line = get_op(line)
        ret += op + indent + line + "\n"
    return ret


def strip_op(text: str) -> str:
    ret: str = ""
    for line in text.split("\n"):
        op, indent, line = get_op(line)
        if op != " ":
            indent = indent[1:]
        ret += indent + line + "\n"
    return ret


def tree_to_diff(diff_tree: ConfigTree) -> Diff:
    ret = []
    for row, v in diff_tree.items():
        op, _, row = get_op(row)
        diff_op = diff_ops[op]
        children = []
        d_match = None
        if isinstance(v, dict):
            children = tree_to_diff(v)
        ret.append((diff_op, row, children, d_match))
    return ret


def diff_to_tree(diff: Diff) -> ConfigTree:
    ret = collections.OrderedDict()
    for diff_op, row, children, _ in diff:
        row = ops_sign[diff_op] + row
        ret[row] = diff_to_tree(children)
    return ret


def _tree_expand_lists_nokia_jun(diff_tree: DiffTree):
    """
    Раскрываем списки Nokia/Juniper в отдельные элементы
    {command: {"[a, b, c]": {}}}   ->   {command a: {}, command b: {}, command c: {}}

    В неупорядоченном множестве префиксов также стираем ';' на конце - их не бывает в патче
    {prefix-list: {"2a02::/64;": {}, "2a03::/64;": {}}}   ->   {prefix-list: {"2a02::/64": {}, "2a03::/64": {}}}
    """
    process: typing.List[DiffTree] = [diff_tree]
    list_regexp = re.compile(r"^(.*)\s+\[(.+)\]$")
    while process:
        tree, process = process[0], process[1:]
        for cmd in list(tree.keys()):
            children = tree[cmd]
            matches = list_regexp.search(cmd)
            normalized_cmd = cmd.rstrip(";")
            if matches:
                for c in matches.group(2).split(" "):
                    if c.strip():
                        tree[" ".join([matches.group(1), c])] = children
            if normalized_cmd != cmd:
                del tree[cmd]
                tree[normalized_cmd] = children
            if isinstance(children, dict):
                process.append(children)



from __future__ import annotations

import abc
import contextlib
import dataclasses
import importlib
import os
import pkgutil
import re
import textwrap
import time
import types
from collections import OrderedDict as odict
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

from annet.annlib import jsontools
from annet.annlib.rbparser.acl import compile_acl_text
from contextlog import get_logger

from annet.storage import Device, Storage

from annet import patching, tabparser, tracing
from annet.cli_args import GenSelectOptions, ShowGeneratorsOptions
from annet.lib import (
    add_annotation,
    flatten,
    get_context,
    jinja_render,
    mako_render,
    merge_dicts,
)
from annet.reference import RefMatcher, RefTracker
from annet.tracing import tracing_connector
from annet.types import (
    GeneratorEntireResult,
    GeneratorJSONFragmentResult,
    GeneratorPartialResult,
    GeneratorPartialRunArgs,
    GeneratorPerf,
    GeneratorResult,
)


# =====
DISABLED_TAG = "disable"


# =====
class GeneratorError(Exception):
    pass


class NotSupportedDevice(GeneratorError):
    pass


class DefaultBlockIfCondition:
    pass


class GeneratorPerfMesurer:
    def __init__(
        self,
        gen: Union["PartialGenerator", "Entire"],
        storage: Storage,
        run_args: Optional[GeneratorPartialRunArgs] = None,
        trace_min_duration: tracing.MinDurationT = None
    ):
        self._gen = gen
        self._storage = storage
        self._run_args = run_args

        self._start_time: float = 0.0
        self._span_ctx = None
        self._span = None
        self._trace_min_duration = trace_min_duration

        self.last_result: Optional[GeneratorPerf] = None

    def start(self) -> None:
        self.last_result = None

        self._storage.flush_perf()

        self._span_ctx = tracing_connector.get().start_as_current_span(
            "gen:call",
            tracer_name=self._gen.__class__.__module__,
            min_duration=self._trace_min_duration,
        )
        self._span = self._span_ctx.__enter__()  # pylint: disable=unnecessary-dunder-call

        if self._span:
            self._span.set_attributes({"generator.class": self._gen.__class__.__name__})
            if self._run_args:
                tracing_connector.get().set_device_attributes(self._span, self._run_args.device)

        self._start_time = time.monotonic()

    def finish(self, exc_type=None, exc_val=None, exc_tb=None) -> GeneratorPerf:
        total = time.monotonic() - self._start_time
        rt = self._storage.flush_perf()
        self._span_ctx.__exit__(exc_type, exc_val, exc_tb)

        meta = {}
        if tracing_connector.get().enabled:
            span_context = self._span.get_span_context()
            meta = {
                "span": {
                    "trace_id": str(span_context.trace_id),
                    "span_id": str(span_context.span_id),
                }
            }

        self.last_result = GeneratorPerf(total=total, rt=rt, meta=meta)
        return self.last_result

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finish(exc_type, exc_val, exc_tb)


class RunGeneratorResult:
    """
    Результат запуска run_partial_generators/run_file_generators
    """

    def __init__(self):
        self.partial_results: Dict[str, GeneratorPartialResult] = {}
        self.entire_results: Dict[str, GeneratorEntireResult] = {}
        self.json_fragment_results: Dict[str, GeneratorJSONFragmentResult] = {}
        self.ref_track: RefTracker = RefTracker()
        self.ref_matcher: RefMatcher = RefMatcher()

    def add_partial(self, result: GeneratorPartialResult):
        self.partial_results[result.name] = result

    def add_entire(self, result: GeneratorEntireResult) -> None:
        # Если есть несколько генераторов на один файл, выбрать тот, что с большим приоритетом
        if result.path:
            if result.path not in self.entire_results or result.prio > self.entire_results[result.path].prio:
                self.entire_results[result.path] = result

    def add_json_fragment(self, result: GeneratorJSONFragmentResult) -> None:
        self.json_fragment_results[result.name] = result

    def config_tree(self, safe: bool = False) -> Dict[str, Any]:  # OrderedDict
        tree = odict()
        for gr in self.partial_results.values():
            config = gr.safe_config if safe else gr.config
            tree = merge_dicts(tree, config)
        return tree

    def new_files(self, safe: bool = False) -> Dict[str, Tuple[str, str]]:
        files = {}
        for gr in self.entire_results.values():
            if not safe or gr.is_safe:
                files[gr.path] = (gr.output, gr.reload)
        return files

    def acl_text(self) -> str:
        return _combine_acl_text(self.partial_results, lambda gr: gr.acl)

    def acl_safe_text(self) -> str:
        return _combine_acl_text(self.partial_results, lambda gr: gr.acl_safe)

    def new_json_fragment_files(
            self,
            old_files: Dict[str, Optional[str]],
            safe: bool = False,
    ) -> Dict[str, Tuple[Any, Optional[str]]]:
        files: Dict[str, Tuple[Any, Optional[str]]] = {}
        reload_prios: Dict[str, int] = {}
        for generator_result in self.json_fragment_results.values():
            filepath = generator_result.path
            if filepath not in files:
                if old_files.get(filepath) is not None:
                    files[filepath] = (old_files[filepath], None)
                else:
                    files[filepath] = ({}, None)
            previous_config: Dict[str, Any] = files[filepath][0]
            new_fragment = generator_result.config

            if safe:
                new_config = jsontools.apply_json_fragment(previous_config, new_fragment, generator_result.acl_safe)
            else:
                new_config = jsontools.apply_json_fragment(previous_config, new_fragment, generator_result.acl)

            if filepath in reload_prios and reload_prios[filepath] > generator_result.reload_prio:
                _, reload_cmd = files[filepath]
            else:
                reload_cmd = generator_result.reload
                reload_prios[filepath] = generator_result.reload_prio
            files[filepath] = (new_config, reload_cmd)
        return files

    def perf_mesures(self) -> Dict[str, Dict[str, int]]:
        mesures = {}
        for gr in self.partial_results.values():
            mesures[gr.name] = {"total": gr.perf.total, "rt": gr.perf.rt, "meta": gr.perf.meta}
        for gr in self.entire_results.values():
            mesures[gr.name] = {"total": gr.perf.total, "rt": gr.perf.rt, "meta": gr.perf.meta}
        return mesures


# =====
def get_list(args: ShowGeneratorsOptions):
    if args.generators_context is not None:
        os.environ["ANN_GENERATORS_CONTEXT"] = args.generators_context
    return {
        cls.__class__.__name__: {
            "type": cls.TYPE,
            "tags": set(cls.TAGS),
            "description": get_description(cls.__class__),
        }
        for cls in _get_generators(get_context()["generators"], None)
    }


def get_description(gen_cls) -> str:
    return textwrap.dedent(" ".join([
        (gen_cls.__doc__ or ""),
        ("Disabled. Use '-g %s' to enable" % gen_cls.__name__ if DISABLED_TAG in gen_cls.TAGS else "")
    ])).strip()


def validate_genselect(gens: GenSelectOptions, all_classes):
    logger = get_logger()
    unknown_err = "Unknown generator alias %s"
    all_aliases = {
        alias
        for cls in all_classes
        for alias in cls.get_aliases()
    }
    for gen_set in (gens.allowed_gens, gens.force_enabled):
        for alias in set(gen_set or ()) - all_aliases:
            logger.error(unknown_err, alias)
            raise Exception(unknown_err % alias)


@dataclasses.dataclass
class Generators:
    """Collection of various types of generators."""

    partial: List[PartialGenerator] = dataclasses.field(default_factory=list)
    entire: List[Entire] = dataclasses.field(default_factory=list)
    json_fragment: List[JSONFragment] = dataclasses.field(default_factory=list)


def build_generators(storage, gens: GenSelectOptions, device: Optional[Device] = None) -> Generators:
    """Return generators that meet the gens filter conditions."""
    if gens.generators_context is not None:
        os.environ["ANN_GENERATORS_CONTEXT"] = gens.generators_context
    all_generators = _get_generators(get_context()["generators"], storage, device)
    validate_genselect(gens, all_generators)
    classes = list(select_generators(gens, all_generators))
    partial = [obj for obj in classes if obj.TYPE == "PARTIAL"]
    entire = [obj for obj in classes if obj.TYPE == "ENTIRE"]
    entire = list(sorted(entire, key=lambda x: x.prio, reverse=True))
    json_fragment = [obj for obj in classes if obj.TYPE == "JSON_FRAGMENT"]
    return Generators(partial=partial, entire=entire, json_fragment=json_fragment)


@tracing.function
def run_partial_initial(device, storage):
    from .common.initial import InitialConfig

    tracing_connector.get().set_device_attributes(tracing_connector.get().get_current_span(), device)

    run_args = GeneratorPartialRunArgs(device, storage)
    return run_partial_generators([InitialConfig()], run_args)


@tracing.function
def run_partial_generators(gens: List["PartialGenerator"], run_args: GeneratorPartialRunArgs):
    logger = get_logger(host=run_args.device.hostname)
    tracing_connector.get().set_device_attributes(tracing_connector.get().get_current_span(), run_args.device)

    ret = RunGeneratorResult()
    if run_args.generators_context is not None:
        os.environ["ANN_GENERATORS_CONTEXT"] = run_args.generators_context
    for gen in _get_ref_generators(get_context()["generators"], run_args.storage):
        ret.ref_matcher.add(gen.ref(run_args.device), gen.__class__)

    logger.debug("Generating selected PARTIALs ...")

    for gen in gens:
        try:
            result = _run_partial_generator(gen, run_args)
        except NotSupportedDevice as exc:
            logger.info("generator %s raised unsupported error: %r", gen, exc)
            continue

        if not result:
            continue

        config = result.safe_config if run_args.use_acl_safe else result.config

        ref_match = ret.ref_matcher.match(config)
        for gen_cls, groups in ref_match:
            gens.append(gen_cls(run_args.storage, groups))
            ret.ref_track.add(gen.__class__, gen_cls)

        ret.ref_track.config(gen.__class__, config)
        ret.add_partial(result)

    return ret


@tracing.function(name="run_partial_generator")
def _run_partial_generator(gen: "PartialGenerator", run_args: GeneratorPartialRunArgs) -> Optional[GeneratorPartialResult]:
    logger = get_logger(generator=_make_generator_ctx(gen))
    device = run_args.device
    output = ""
    config = odict()
    safe_config = odict()

    if not gen.supports_device(device):
        logger.info("generator %s is not supported for device %s", gen, device.hostname)
        return None

    span = tracing_connector.get().get_current_span()
    if span:
        tracing_connector.get().set_device_attributes(span, run_args.device)
        tracing_connector.get().set_dimensions_attributes(span, gen, run_args.device)
        span.set_attributes({
            "use_acl": run_args.use_acl,
            "use_acl_safe": run_args.use_acl_safe,
            "generators_context": str(run_args.generators_context),
        })

    with GeneratorPerfMesurer(gen, run_args.storage, run_args=run_args) as pm:
        if not run_args.no_new:
            if gen.get_user_runner(device):
                logger.info("Generating PARTIAL ...")
            try:
                output = gen(device, run_args.annotate)
            except NotSupportedDevice:
                # это исключение нужно передать выше в оригинальном виде
                raise
            except Exception as err:
                filename, lineno = gen.get_running_line()
                logger.exception("Generator error in file '%s:%i'", filename, lineno)
                raise GeneratorError(f"{gen} on {device}") from err

            fmtr = tabparser.make_formatter(device.hw)
            try:
                config = tabparser.parse_to_tree(text=output, splitter=fmtr.split)
            except tabparser.ParserError as err:
                logger.exception("Parser error")
                raise GeneratorError from err

    acl = gen.acl(device) or ""
    rules = compile_acl_text(textwrap.dedent(acl), device.hw.vendor)
    acl_safe = gen.acl_safe(device) or ""
    safe_rules = compile_acl_text(textwrap.dedent(acl_safe), device.hw.vendor)

    if run_args.use_acl:
        try:
            with tracing_connector.get().start_as_current_span("apply_acl", tracer_name=__name__, min_duration="0.01") as acl_span:
                tracing_connector.get().set_device_attributes(acl_span, run_args.device)
                config = patching.apply_acl(
                    config=config,
                    rules=rules,
                    fatal_acl=True,
                    with_annotations=run_args.annotate,
                )
            if run_args.use_acl_safe:
                with tracing_connector.get().start_as_current_span(
                    "apply_acl_safe",
                    tracer_name=__name__,
                    min_duration="0.01"
                ) as acl_safe_span:
                    tracing_connector.get().set_device_attributes(acl_safe_span, run_args.device)
                    safe_config = patching.apply_acl(
                        config=config,
                        rules=safe_rules,
                        fatal_acl=False,
                        with_annotations=run_args.annotate,
                    )
        except patching.AclError as err:
            logger.error("ACL error: generator is not allowed to yield this command: %s", err)
            raise GeneratorError from err
        except NotImplementedError as err:
            logger.error(str(err))
            raise GeneratorError from err

    return GeneratorPartialResult(
        name=gen.__class__.__name__,
        tags=gen.TAGS,
        output=output,
        acl=acl,
        acl_rules=rules,
        acl_safe=acl_safe,
        acl_safe_rules=safe_rules,
        config=config,
        safe_config=safe_config,
        perf=pm.last_result,
    )


@tracing.function
def check_entire_generators_required_packages(gens, device_packages: FrozenSet[str]) -> List[str]:
    errors: List[str] = []
    for gen in gens:
        if not gen.REQUIRED_PACKAGES.issubset(device_packages):
            missing = gen.REQUIRED_PACKAGES - device_packages
            missing_str = ", ".join("`{}'".format(pkg) for pkg in sorted(missing))
            if len(missing) == 1:
                errors.append("missing package {} required for {}".format(missing_str, gen))
            else:
                errors.append("missing packages {} required for {}".format(missing_str, gen))
    return errors


@tracing.function
def run_file_generators(
        gens: Iterable[Union["JSONFragment", "Entire"]],
        device: "Device",
        storage: Storage,
) -> RunGeneratorResult:
    """Run generators that generate files or file parts."""
    ret = RunGeneratorResult()
    logger = get_logger(host=device.hostname)
    logger.debug("Generating selected ENTIREs and JSON_FRAGMENTs ...")
    for gen in gens:
        if gen.__class__.TYPE == "ENTIRE":
            run_generator_fn = _run_entire_generator
            add_result_fn = ret.add_entire
        elif gen.__class__.TYPE == "JSON_FRAGMENT":
            run_generator_fn = _run_json_fragment_generator
            add_result_fn = ret.add_json_fragment
        else:
            raise RuntimeError(f"Unknown generator class type: cls={gen.__class__} TYPE={gen.__class__.TYPE}")
        try:
            result = run_generator_fn(gen, device, storage)
        except NotSupportedDevice as exc:
            logger.info("generator %s raised unsupported error: %r", gen, exc)
            continue
        if result:
            add_result_fn(result)

    return ret


@tracing.function(min_duration="0.5")
def _run_entire_generator(gen: "Entire", device: "Device", storage: Storage) -> Optional[GeneratorResult]:
    logger = get_logger(generator=_make_generator_ctx(gen))
    if not gen.supports_device(device):
        logger.info("generator %s is not supported for device %s", gen, device.hostname)
        return

    span = tracing_connector.get().get_current_span()
    if span:
        tracing_connector.get().set_device_attributes(span, device)
        tracing_connector.get().set_dimensions_attributes(span, gen, device)

    path = gen.path(device)
    if not path:
        raise RuntimeError("entire generator should return non-empty path")

    logger.info("Generating ENTIRE ...")
    with GeneratorPerfMesurer(gen, storage, trace_min_duration="0.5") as pm:
        output = gen(device)

    return GeneratorEntireResult(
        name=gen.__class__.__name__,
        tags=gen.TAGS,
        path=path,
        output=output,
        reload=gen.get_reload_cmds(device),
        prio=gen.prio,
        perf=pm.last_result,
        is_safe=gen.is_safe(device),
    )


def _make_generator_ctx(gen):
    return "%s.[%s]" % (gen.__module__, gen.__class__.__name__)


def _run_json_fragment_generator(
        gen: "JSONFragment",
        device: "Device",
        storage: Storage,
) -> Optional[GeneratorResult]:
    logger = get_logger(generator=_make_generator_ctx(gen))
    if not gen.supports_device(device):
        logger.info("generator %s is not supported for device %s", gen, device.hostname)
        return

    path = gen.path(device)
    if not path:
        raise RuntimeError("json fragment generator should return non-empty path")

    acl_item_or_list_of_items = gen.acl(device)
    safe_acl_item_or_list_of_items = gen.acl_safe(device)
    if not acl_item_or_list_of_items:
        raise RuntimeError("json fragment generator should return non-empty acl")
    if isinstance(acl_item_or_list_of_items, list):
        acl = acl_item_or_list_of_items
    else:
        acl = [acl_item_or_list_of_items]
    if isinstance(safe_acl_item_or_list_of_items, list):
        acl_safe = safe_acl_item_or_list_of_items
    else:
        acl_safe = [safe_acl_item_or_list_of_items]

    logger.info("Generating JSON_FRAGMENT ...")
    with GeneratorPerfMesurer(gen, storage) as pm:
        config = gen(device)
    reload_cmds = gen.get_reload_cmds(device)
    return GeneratorJSONFragmentResult(
        name=gen.__class__.__name__,
        tags=gen.TAGS,
        path=path,
        acl=acl,
        acl_safe=acl_safe,
        config=config,
        reload=reload_cmds,
        perf=pm.last_result,
        reload_prio=gen.reload_prio,
    )


def _get_generators(module_paths: Union[List[str], dict], storage, device=None):
    if isinstance(module_paths, dict):
        if device is None:
            module_paths = module_paths.get("default")
        else:
            modules = []
            seen = set()
            for prop, prop_modules in module_paths.get("per_device_property", {}).items():
                if getattr(device, prop, False) is True:
                    for module in prop_modules:
                        if module not in seen:
                            modules.append(module)
                            seen.add(module)
            module_paths = modules or module_paths.get("default")
    res_generators = []
    for module_path in module_paths:
        module = importlib.import_module(module_path)
        if hasattr(module, "get_generators"):
            generators: List[BaseGenerator] = module.get_generators(storage)
            res_generators += generators
    return res_generators


def _get_ref_generators(module_paths: List[str], storage):
    if isinstance(module_paths, dict):
        module_paths = module_paths.get("default")
    res_generators = []
    for module_path in module_paths:
        module = importlib.import_module(module_path)
        if hasattr(module, "get_ref_generators"):
            res_generators += module.get_ref_generators(storage)
    return res_generators


class InvalidValueFromGenerator(ValueError):
    pass


class GenStringable(abc.ABC):
    @abc.abstractmethod
    def gen_str(self) -> str:
        pass


ParamsList = tabparser.JuniperList


def _filter_str(value: Union[str, int, float, tabparser.JuniperList, ParamsList, GenStringable]):
    if isinstance(value, (
        str,
        int,
        float,
        tabparser.JuniperList,
        ParamsList,
    )):
        return str(value)

    if hasattr(value, "gen_str") and callable(value.gen_str):
        return value.gen_str()

    raise InvalidValueFromGenerator("Invalid yield type: %s(%s)" % (type(value).__name__, value))


# =====
class BaseGenerator:
    TYPE: str
    TAGS: list[str]

    def supports_device(self, device: Device) -> bool:  # pylint: disable=unused-argument
        return True


class TreeGenerator(BaseGenerator):
    def __init__(self, indent="  "):
        self._indents = []
        self._rows = []
        self._block_path = []
        self._indent = indent

    @tracing.contextmanager(min_duration="0.1")
    @contextlib.contextmanager
    def block(self, *tokens, indent=None):
        span = tracing_connector.get().get_current_span()
        if span:
            span.set_attribute("tokens", " ".join(map(str, tokens)))

        indent = self._indent if indent is None else indent
        block = " ".join(map(_filter_str, tokens))
        self._block_path.append(block)
        self._append_text(block)
        self._indents.append(indent)
        yield
        self._indents.pop(-1)
        self._block_path.pop(-1)

    @contextlib.contextmanager
    def block_if(self, *tokens, condition=DefaultBlockIfCondition):
        if condition is DefaultBlockIfCondition:
            condition = (None not in tokens and "" not in tokens)
        if condition:
            with self.block(*tokens):
                yield
                return
        yield

    @contextlib.contextmanager
    def multiblock(self, *blocks):
        if blocks:
            blk = blocks[0]
            tokens = blk if isinstance(blk, (list, tuple)) else [blk]
            with self.block(*tokens):
                with self.multiblock(*blocks[1:]):
                    yield
                    return
        yield

    @contextlib.contextmanager
    def multiblock_if(self, *blocks, condition=DefaultBlockIfCondition):
        if condition is DefaultBlockIfCondition:
            condition = (None not in blocks)
            if condition:
                if blocks:
                    blk = blocks[0]
                    tokens = blk if isinstance(blk, (list, tuple)) else [blk]
                    with self.block(*tokens):
                        with self.multiblock(*blocks[1:]):
                            yield
                            return
        yield

    # ===
    def _append_text(self, text):
        self._append_text_cb(text)

    def _append_text_cb(self, text, row_cb=None):
        for row in _split_and_strip(text):
            if row_cb:
                row = row_cb(row)
            self._rows.append("".join(self._indents) + row)


class TextGenerator(TreeGenerator):
    def __add__(self, line):
        self._append_text(line)
        return self

    def __iter__(self):
        yield from self._rows


class PartialGenerator(TreeGenerator):
    TYPE = "PARTIAL"
    TAGS: List[str] = []

    def __init__(self, storage):
        super().__init__()
        self.storage = storage
        self._annotate = False
        self._running_gen = None
        self._annotations = []
        self._annotation_module = self.__class__.__module__ or ""

    def supports_device(self, device: Device) -> bool:
        if self.__class__.run is PartialGenerator.run:
            return bool(self._get_vendor_func(device.hw.vendor, "run"))
        else:
            return True

    def acl(self, device):
        acl_func = self._get_vendor_func(device.hw.vendor, "acl")
        if acl_func:
            return acl_func(device)

    def acl_safe(self, device):
        acl_func = self._get_vendor_func(device.hw.vendor, "acl_safe")
        if acl_func:
            return acl_func(device)

    def run(self, device) -> Iterable[Union[str, tuple]]:
        run_func = self._get_vendor_func(device.hw.vendor, "run")
        if run_func:
            return run_func(device)

    def get_user_runner(self, device):
        if self.__class__.run is not PartialGenerator.run:
            return self.run
        return self._get_vendor_func(device.hw.vendor, "run")

    def _get_vendor_func(self, vendor: str, func_name: str):
        attr_name = f"{func_name}_{vendor}"
        if hasattr(self, attr_name):
            return getattr(self, attr_name)
        return None

    # =====

    @classmethod
    def get_aliases(cls) -> Set[str]:
        return {cls.__name__, *cls.TAGS}

    def __call__(self, device, annotate=False):
        self._indents = []
        self._rows = []
        self._running_gen = self.run(device)
        self._annotate = annotate

        if annotate and self.__class__.__module__:
            self._annotation_module = ".".join(self.__class__.__module__.split(".")[-2:])

        for text in self._running_gen:
            if isinstance(text, tuple):
                text = " ".join(map(_filter_str, flatten(text)))
            else:
                text = _filter_str(text)
            self._append_text(text)

        for row in self._rows:
            assert re.search(r"\bNone\b", row) is None, "Found 'None' in yield result: %s" % (row)
        if annotate:
            generated_rows = (add_annotation(x, y) for (x, y) in zip(self._rows, self._annotations))
        else:
            generated_rows = self._rows
        return "\n".join(generated_rows) + "\n"

    def _append_text(self, text):
        def annotation_cb(row):
            annotation = "%s:%d" % self.get_running_line()
            self._annotations.append(annotation)
            return row

        self._append_text_cb(
            text,
            annotation_cb if self._annotate else None
        )

    def get_running_line(self):
        if not self._running_gen or not self._running_gen.gi_frame:
            return (repr(self._running_gen), -1)
        return self._annotation_module, self._running_gen.gi_frame.f_lineno

    @classmethod
    def literal(cls, item):
        return '"{}"'.format(item)

    def __repr__(self):
        return "<%s>" % self.__class__.__name__


class RefGenerator(PartialGenerator):
    def __init__(self, storage, groups=None):
        super().__init__(storage)
        self.groups = groups

    def ref(self, device):
        if hasattr(self, "ref_" + device.hw.vendor):
            return getattr(self, "ref_" + device.hw.vendor)(device)
        return ""


class Entire(BaseGenerator):
    TYPE = "ENTIRE"
    TAGS: List[str] = []
    REQUIRED_PACKAGES: FrozenSet[str] = frozenset()

    def __init__(self, storage):
        self.storage = storage
        # между генераторами для одного и того же path - выбирается тот что больше
        if not hasattr(self, "prio"):
            self.prio = 100
        self.__device = None

    def supports_device(self, device: Device):
        return bool(self.path(device))

    def run(self, device) -> Union[None, str, Iterable[Union[str, tuple]]]:
        raise NotImplementedError

    def reload(self, device) -> Optional[str]:  # pylint: disable=unused-argument
        return

    def get_reload_cmds(self, device) -> str:
        ret = self.reload(device) or ""
        path = self.path(device)
        if path and device.hw.PC and device.hw.soft.startswith(("Cumulus", "SwitchDev", "SONiC")):
            parts = []
            if ret:
                parts.append(ret)
            parts.append("/usr/bin/etckeeper commitreload %s" % path)
            return "\n".join(parts)
        return ret

    def path(self, device) -> Optional[str]:
        raise NotImplementedError("Required PATH for ENTIRE generator")

    # pylint: disable=unused-argument
    def is_safe(self, device) -> bool:
        """Output gen results when --acl-safe flag is used"""
        return False

    def read(self, path) -> str:
        return pkgutil.get_data(__name__, path).decode()

    def mako(self, text, **kwargs) -> str:
        return mako_render(text, dedent=True, device=self.__device, **kwargs)

    def jinja(self, text, **kwargs) -> str:
        return jinja_render(text, dedent=True, device=self.__device, **kwargs)

    # =====

    @classmethod
    def get_aliases(cls) -> Set[str]:
        return {cls.__name__, *cls.TAGS}

    def __call__(self, device):
        self.__device = device
        parts = []
        run_res = self.run(device)
        if isinstance(run_res, str):
            run_res = (run_res,)
        if run_res is None or not isinstance(run_res, (tuple, types.GeneratorType)):
            raise Exception("generator %s returns %s" % (self.__class__.__name__, type(run_res)))
        for text in run_res:
            if isinstance(text, tuple):
                text = " ".join(map(_filter_str, flatten(text)))
            assert re.search(r"\bNone\b", text) is None, "Found 'None' in yield result: %s" % text
            parts.append(text)
        return "\n".join(parts)


def select_generators(gens: GenSelectOptions, classes: Iterable[BaseGenerator]):
    def contains(obj, where):
        if where:
            return obj.get_aliases().intersection(where)
        return False

    def has(cls, what):
        return what in cls.TAGS

    flts = [lambda c: not isinstance(c, RefGenerator)]
    if gens.allowed_gens:
        flts.append(lambda c: contains(c, gens.allowed_gens))
    elif gens.force_enabled:
        flts.append(lambda c: not has(c, DISABLED_TAG) or contains(c, gens.force_enabled))
    elif not gens.ignore_disabled:
        flts.append(lambda c: not has(c, DISABLED_TAG))

    if gens.excluded_gens:
        flts.append(lambda c: not contains(c, gens.excluded_gens))

    return filter(lambda x: all(f(x) for f in flts), classes)


def _split_and_strip(text):
    if "\n" in text:
        rows = textwrap.dedent(text).strip().split("\n")
    else:
        rows = [text]
    return rows


def _combine_acl_text(
    partial_results: Dict[str, GeneratorPartialResult],
    acl_getter: Callable[[GeneratorPartialResult], str]
) -> str:
    acl_text = ""
    for gr in partial_results.values():
        for line in textwrap.dedent(acl_getter(gr)).split("\n"):
            if line and not line.isspace():
                acl_text += line.rstrip()
                acl_text += fr"  %generator_names={gr.name}"
                acl_text += "\n"
    return acl_text


class JSONFragment(TreeGenerator):
    """Generates parts of JSON config file."""

    TYPE = "JSON_FRAGMENT"
    TAGS: List[str] = []

    def __init__(self, storage: Storage):
        super().__init__()
        self.storage = storage
        self._json_config: Dict[str, Any] = {}
        self._config_pointer: List[str] = []

        # if two generators edit same file, commands from generator with greater `reload_prio` will be used
        if not hasattr(self, "reload_prio"):
            self.reload_prio = 100

    def supports_device(self, device: Device):
        return bool(self.path(device))

    def path(self, device: Device) -> Optional[str]:
        raise NotImplementedError("Required PATH for JSON_FRAGMENT generator")

    @classmethod
    def get_aliases(cls) -> Set[str]:
        return {cls.__name__, *cls.TAGS}

    def acl(self, device: Device) -> Union[str, List[str]]:
        """
        Restrict the generator to a specified ACL using JSON Pointer syntax.

        Expected ACL to be a list of strings, but a single string is also allowed.
        """
        raise NotImplementedError("Required ACL for JSON_FRAGMENT generator")

    def acl_safe(self, device: Device) -> Union[str, List[str]]:
        """
        Restrict the generator to a specified safe ACL using JSON Pointer syntax.

        Expected ACL to be a list of strings, but a single string is also allowed.
        """
        raise NotImplementedError("Required ACL for JSON_FRAGMENT generator")

    def run(self, device: Device):
        raise NotImplementedError

    def get_reload_cmds(self, device: Device) -> str:
        ret = self.reload(device) or ""
        return ret

    def reload(self, device) -> Optional[str]:
        raise NotImplementedError

    @contextlib.contextmanager
    def block(self, *tokens, indent=None):  # pylint: disable=unused-argument
        block_str = "".join(map(_filter_str, tokens))
        self._config_pointer.append(block_str)
        try:
            yield
        finally:
            self._config_pointer.pop()

    @contextlib.contextmanager
    def block_piped(self, *tokens, indent=None):  # pylint: disable=unused-argument
        block_str = "|".join(map(_filter_str, tokens))
        self._config_pointer.append(block_str)
        try:
            yield
        finally:
            self._config_pointer.pop()

    def __call__(self, device: Device, annotate: bool = False):
        for cfg_fragment in self.run(device):
            self._set_or_replace_dict(self._config_pointer, cfg_fragment)
        return self._json_config

    def _set_or_replace_dict(self, pointer, value):
        if not pointer:
            if self._json_config == {}:
                self._json_config = value
            else:
                self._json_config = [self._json_config, value]
        else:
            self._set_dict(self._json_config, pointer, value)

    @classmethod
    def _to_str(cls, value: Any) -> str:
        if isinstance(value, str):
            return value
        elif isinstance(value, list):
            return [cls._to_str(x) for x in value]
        elif isinstance(value, dict):
            for k, v in value.items():
                value[k] = cls._to_str(v)
            return value
        return str(value)

    @classmethod
    def _set_dict(cls, cfg, pointer, value):
        # pointer has at least one key
        if len(pointer) == 1:
            if pointer[0] in cfg:
                cfg[pointer[0]] = [cfg[pointer[0]], cls._to_str(value)]
            else:
                cfg[pointer[0]] = cls._to_str(value)
        else:
            if pointer[0] not in cfg:
                cfg[pointer[0]] = {}
            cls._set_dict(cfg[pointer[0]], pointer[1:], cls._to_str(value))



import functools
import re
from collections import OrderedDict as odict

from annet.annlib.rbparser import platform, syntax
from valkit.common import valid_bool, valid_string_list
from valkit.python import valid_object_path

from .common import import_rulebook_function


DEFAULT_PATCH_LOGIC = "common.default"
ORDERED_PATCH_LOGIC = "common.ordered"
REWRITE_PATCH_LOGIC = "common.rewrite"
REWRITE_DIFF_LOGIC = "common.rewrite_diff"
MULTILINE_DIFF_LOGIC = "common.multiline_diff"


# =====
@functools.lru_cache()
def compile_patching_text(text, vendor):
    return _compile_patching(
        tree=syntax.parse_text(text, params_scheme={
            "global": {
                "validator": valid_bool,
                "default": False,
            },
            "logic": {
                "validator": valid_object_path,
                "default": DEFAULT_PATCH_LOGIC,
            },
            "diff_logic": {
                "validator": valid_object_path,
                "default": platform.VENDOR_DIFF[vendor],
            },
            "comment": {
                "validator": valid_string_list,
                "default": [],
            },
            "multiline": {
                "validator": valid_bool,
                "default": False,
            },
            "ordered": {
                "validator": valid_bool,
                "default": False,
            },
            "rewrite": {
                "validator": valid_bool,
                "default": False,
            },
            "parent": {
                "validator": valid_bool,
                "default": False,
            },
            "force_commit": {
                "validator": valid_bool,
                "default": False,
            },
            "ignore_case": {
                "validator": valid_bool,
                "default": False,
            },
        }),
        reverse_prefix=platform.VENDOR_REVERSES[vendor],
        vendor=vendor,
    )


# =====
def _compile_patching(tree, reverse_prefix, vendor):
    rules = {"local": odict(), "global": odict()}
    for (raw_rule, attrs) in tree.items():
        regexp = _attrs_to_regexp(attrs)
        attrs = _regexp_to_attrs(regexp, attrs)
        if attrs["type"] == "ignore":
            rule = {
                "type": attrs["type"],
                "attrs": {
                    "regexp": regexp,
                    "diff_logic": import_rulebook_function(attrs["params"]["diff_logic"]),
                    "parent": bool(attrs["children"]),
                    "context": attrs["context"],
                },
                "children": {"global": {}, "local": {}}
            }
        else:
            if attrs["params"]["ordered"]:
                attrs["params"]["diff_logic"] = platform.VENDOR_DIFF_ORDERED[vendor]
                attrs["params"]["logic"] = ORDERED_PATCH_LOGIC
            elif attrs["params"]["rewrite"]:
                attrs["params"]["diff_logic"] = REWRITE_DIFF_LOGIC
                attrs["params"]["logic"] = REWRITE_PATCH_LOGIC
            elif attrs["params"]["multiline"]:
                attrs["params"]["diff_logic"] = MULTILINE_DIFF_LOGIC
            rule = {
                "type": attrs["type"],
                "attrs": {
                    "logic": import_rulebook_function(attrs["params"]["logic"]),
                    "diff_logic": import_rulebook_function(attrs["params"]["diff_logic"]),
                    "regexp": regexp,
                    "reverse": _make_reverse(attrs["row"], reverse_prefix, flags=regexp.flags),
                    "comment": attrs["params"]["comment"],
                    "multiline": attrs["params"]["multiline"],
                    "parent": attrs["params"]["parent"] or bool(attrs["children"]),
                    "force_commit": attrs["params"]["force_commit"],
                    "ignore_case": attrs["params"]["ignore_case"],
                    "context": attrs["context"],
                },
                "children": None,
            }
            if not attrs["params"]["global"]:
                rule["children"] = _compile_patching(attrs["children"], reverse_prefix, vendor)
        rules["global" if attrs["params"]["global"] else "local"][raw_rule] = rule
    return rules


@functools.lru_cache()
def _make_reverse(row, reverse_prefix, flags=0):
    if row.startswith(reverse_prefix + " "):
        row = row[len(reverse_prefix + " "):]
    else:
        row = "%s %s" % (reverse_prefix, row)

    if row[-1] == "~":
        row = row[:-1] + "{}"

    row = re.sub(r"\*(/\S+/)?", "{}", row, flags=flags)
    row = re.sub(r"\s*~(/\S+/)?", "", row, flags=flags)
    return row


def _attrs_to_regexp(attrs):
    flags = 0
    ignore_case = attrs["params"]["ignore_case"]
    if ignore_case:
        flags |= re.IGNORECASE
    return syntax.compile_row_regexp(attrs["row"], flags=flags)


def _regexp_to_attrs(regexp, attrs):
    attrs["params"]["ignore_case"] = bool(regexp.flags & re.IGNORECASE)
    return attrs



from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional, Any

from annet.annlib.netdev.views.hardware import HardwareView


@dataclass
class Entity:
    id: int
    name: str


@dataclass
class Label:
    value: str
    label: str


@dataclass
class IpFamily:
    value: int
    label: str


@dataclass
class DeviceType:
    id: int
    manufacturer: Entity
    model: str


@dataclass
class DeviceIp:
    id: int
    display: str
    address: str
    family: int


@dataclass
class IpAddress:
    id: int
    assigned_object_id: int
    display: str
    family: IpFamily
    address: str
    status: Label
    tags: List[Entity]
    created: datetime
    last_updated: datetime


@dataclass
class Interface(Entity):
    device: Entity
    enabled: bool
    display: str = ""
    ip_addresses: List[IpAddress] = field(default_factory=list)


@dataclass
class NetboxDevice(Entity):
    neighbours_ids: List[int]

    display: str
    device_type: DeviceType
    device_role: Entity
    tenant: Optional[Entity]
    platform: Optional[Entity]
    serial: str
    asset_tag: Optional[str]
    site: Entity
    rack: Optional[Entity]
    position: Optional[float]
    face: Optional[Label]
    status: Label
    primary_ip: Optional[DeviceIp]
    primary_ip4: Optional[DeviceIp]
    primary_ip6: Optional[DeviceIp]
    tags: List[Entity]
    custom_fields: dict[str, Any]
    created: datetime
    last_updated: datetime

    fqdn: str
    hostname: str
    hw: Optional[HardwareView]
    breed: str

    interfaces: List[Interface]

    # compat
    def __hash__(self):
        return hash(self.id)

    def is_pc(self):
        return self.device_type.manufacturer.name == "Mellanox"



from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional, Any

from annet.adapters.netbox.common.models import (
    Entity, Label, DeviceType, DeviceIp,
)


@dataclass
class Interface(Entity):
    device: Entity
    enabled: bool
    display: str = ""  # added in 3.x


@dataclass
class Device(Entity):
    display: str  # renamed in 3.x from display_name
    device_type: DeviceType
    device_role: Entity
    tenant: Optional[Entity]
    platform: Optional[Entity]
    serial: str
    asset_tag: Optional[str]
    site: Entity
    rack: Optional[Entity]
    position: Optional[float]
    face: Optional[Label]
    status: Label
    primary_ip: Optional[DeviceIp]
    primary_ip4: Optional[DeviceIp]
    primary_ip6: Optional[DeviceIp]
    tags: List[Entity]
    custom_fields: dict[str, Any]
    created: datetime
    last_updated: datetime



from datetime import datetime
from typing import List, Optional

import dateutil.parser
from adaptix import Retort, loader
from dataclass_rest import get

from annet.adapters.netbox.common.client import (
    BaseNetboxClient, collect, PagingResponse,
)
from .api_models import Device, Interface, IpAddress


class NetboxV24(BaseNetboxClient):
    def _init_response_body_factory(self) -> Retort:
        return Retort(recipe=[
            loader(datetime, dateutil.parser.parse)
        ])

    @get("dcim/interfaces")
    def interfaces(
            self,
            device_id: Optional[List[int]] = None,
            limit: int = 20,
            offset: int = 0,
    ) -> PagingResponse[Interface]:
        pass

    all_interfaces = collect(interfaces, field="device_id")

    @get("ipam/ip-addresses")
    def ip_addresses(
            self,
            interface_id: Optional[List[int]] = None,
            limit: int = 20,
            offset: int = 0,
    ) -> PagingResponse[IpAddress]:
        pass

    all_ip_addresses = collect(ip_addresses, field="interface_id")

    @get("dcim/devices")
    def devices(
            self,
            name: Optional[List[str]] = None,
            tag: Optional[str] = None,
            limit: int = 20,
            offset: int = 0,
    ) -> PagingResponse[Device]:
        pass

    all_devices = collect(devices)

    @get("dcim/devices/{device_id}")
    def get_device(
            self,
            device_id: int,
    ) -> Device:
        pass



import inspect
import sys
import traceback


class DumpResult(str):
    """
    строка, у которой repr = value
    """
    def __new__(cls, value):
        return super().__new__(cls, value)

    __repr__ = str.__str__


class _EnumAllAttrs:
    pass


class DumpableView:
    def _enum_attrs(self):
        for attr in dir(self):
            if not attr.startswith("_") and attr != "dump":
                yield attr

    def _getter_for_dump(self, attr):
        dumper_name = "_dump_" + attr
        if hasattr(self, dumper_name) and callable(getattr(self, dumper_name)):
            return getattr(self, dumper_name)

    @staticmethod
    def strip_dumper(attr, lines=None, chars=None, chars_in_line=None):
        def bound_method(self):
            lines_list = []
            value = getattr(self, attr)
            assert isinstance(value, str)
            total_lines = 0
            total_chars = 0
            add_tail = True
            for line in value.split("\n"):
                if lines and total_lines >= lines:
                    break

                if chars and total_chars + len(line) >= chars:
                    lines_list.append(line[chars - total_chars:])
                    break

                if chars_in_line and len(line) > chars_in_line:
                    lines_list.append(line[chars - total_chars:] + " ...")
                else:
                    lines_list.append(line)
                total_lines += 1
                total_chars += len(line) + 1
            else:
                add_tail = False

            ret = repr("\n".join(lines_list))
            if add_tail:
                ret += " ..."
            return DumpResult(ret)
        return bound_method

    def __dump_value(self, prefix, value, seen):
        ret = []
        op = "="
        if isinstance(value, DumpableView):
            # защита от рекурсии - последующие ссылки на объекты, а не новые дампы
            if id(value) in seen:
                value = DumpResult(seen[id(value)])
                op = "->"
            else:
                seen[id(value)] = prefix

        if isinstance(value, DumpableView):
            ret += value.dump(prefix, seen=seen)  # pylint: disable=no-member
        else:
            fmt = repr(value)
            vtype = type(value)
            if not isinstance(value, DumpResult) and (vtype.__module__ != "builtins") and vtype.__name__ not in fmt:
                fmt += "  # %s" % (vtype)
            ret += ["%s %s %s" % (prefix, op, fmt)]
        return ret

    def dump(self, prefix="", value=_EnumAllAttrs, seen=None):
        """
        В собственных DumpableView-классах нужно определить dump примерно так:
        def dump(self, prefix, **kwargs):
            ret = super().dump(prefix, **kwargs) + [ your_own_lines ]
        """
        if seen is None:
            seen = {id(self): prefix}

        ret = []
        if value is not _EnumAllAttrs:
            ret.extend(self.__dump_value(prefix, value, seen))
        else:
            for attr in self._enum_attrs():
                getter = self._getter_for_dump(attr)
                try:
                    if getter:
                        value = getter()
                    else:
                        value = getattr(self, attr)
                        if inspect.isroutine(value):
                            continue
                except Exception as exc:
                    exc_tb = sys.exc_info()[2]
                    frame_index = -1
                    tb_list = traceback.extract_tb(exc_tb)
                    value = DumpResult("%r at %s:%s" % (exc, tb_list[frame_index][0], tb_list[frame_index][1]))

                ret.extend(self.__dump_value("%s.%s" % (prefix, attr), value, seen))
        return ret

    # Filter out cached lru_methods from state
    def __getstate__(self):
        state = self.__dict__.copy()
        for key in list(state.keys()):  # For thread-safe: dictionary changed size during iteration
            if inspect.ismethod(state[key]):
                del state[key]
        return state



import abc
import os
import posixpath
import sys
from typing import List, Optional, Tuple, Type

import colorama
from annet.annlib.output import (  # pylint: disable=unused-import
    LABEL_NEW_PREFIX,
    OutputWriter,
    TextArgs,
    capture_output,
    dir_or_file_output,
    format_file_diff,
    print_as_json,
    print_as_yaml,
    print_err_label,
    print_label,
)
from contextlog import get_logger

from annet.cli_args import FileOutOptions, QueryOptions
from annet.connectors import Connector
from annet.storage import Device, storage_connector


class _DriverConnector(Connector["OutputDriver"]):
    name = "OutputDriver"
    ep_name = "output"

    def _get_default(self) -> Type["OutputDriver"]:
        return OutputDriverBasic


output_driver_connector = _DriverConnector()


class OutputDriver(abc.ABC):
    @abc.abstractmethod
    def write_output(self, arg_out: FileOutOptions, items, query_result_count=1) -> None:
        pass

    @abc.abstractmethod
    def format_fails(self, fail, args: Optional[QueryOptions] = None) -> Tuple[str, str]:
        pass

    @abc.abstractmethod
    def cfg_file_names(self, device: Device) -> List[str]:
        pass

    @abc.abstractmethod
    def entire_config_dest_path(self, device: Device, config_path: str) -> str:
        pass


class OutputDriverBasic(OutputDriver):
    def write_output(self, arg_out: FileOutOptions, items, query_result_count=1):
        """
        пишет результаты генерации в файл или директорию :dest
        :dest - это директория в случаях:
        - заканчивается на "/"
        - существует директория с таким именем
        - более одного устройства в результате запроса
        - есть entire-генераторы (определяется по типу первого результата, если устройство одно)
        """
        logger = get_logger()

        items_iter = iter(items)
        try:
            first_result = next(items_iter)
        except StopIteration:
            # нет результатов, ничего не пишем и не создаём
            return

        def _reassemble_items():
            yield first_result
            yield from items_iter

        dest = arg_out.dest
        dir_mode = dir_or_file_output(dest, query_result_count, suggest_dir=(os.sep in first_result[0]))
        _reassembled_items = list(_reassemble_items())

        for output_no, (label, output, is_fail) in enumerate(_reassembled_items):
            writer = output if isinstance(output, OutputWriter) else OutputWriter(output)
            label = os.path.normpath(label)
            label_color = colorama.Back.RED if is_fail else colorama.Back.GREEN
            if dest is None:
                if hasattr(arg_out, "format") and arg_out.format == "json":
                    if output_no > 0:
                        sys.stdout.write(",")
                    elif output_no == 0:
                        sys.stdout.write("{")
                    sys.stdout.write('"%s": ' % label)
                    writer.write(sys.stdout)
                    if output_no == len(_reassembled_items) - 1:
                        sys.stdout.write("}")
                else:
                    if not arg_out.no_label:
                        print_label(label, back_color=label_color)
                    writer.write(sys.stdout)
            elif dir_mode:
                if label.startswith(LABEL_NEW_PREFIX):
                    label = label[len(LABEL_NEW_PREFIX):]
                if label.startswith(os.sep):
                    # just in case.
                    label = label.lstrip(os.sep)
                if os.sep not in label:
                    # vendor config
                    label = os.path.basename(label)
                else:
                    # entire generated file
                    parts = label.split(os.sep)
                    hostname = parts[0]
                    label = os.sep.join(parts[1:])
                    if not arg_out.expand_path:
                        label = os.path.basename(label)
                    if query_result_count > 1:
                        label = os.path.join(hostname, label)
                file_dest = os.path.join(dest, "errors") if is_fail else dest
                out_file = os.path.normpath(os.path.join(file_dest, label))
                logger.info("writing '%s'", out_file)
                dirname = os.path.dirname(out_file)
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
                with open(out_file, "w") as file:
                    writer.write(file)
            else:
                logger.info("writing '%s'", dest)
                with open(dest, "w") as file:
                    writer.write(file)

    def format_fails(self, fail, args: Optional[QueryOptions] = None):
        ret = []
        fqdns = {}
        if args:
            connector = storage_connector.get()
            storage_opts = connector.opts().from_cli_opts(args)
            with connector.storage()(storage_opts) as storage:
                fqdns = storage.resolve_fdnds_by_query(args.query)
        for (assignment, exc) in fail.items():
            label = assignment
            if assignment in fqdns:
                label = fqdns[assignment]
            elif isinstance(assignment, tuple):
                label = assignment[0]
            else:
                ValueError("Failed to parse failed assignment %r" % assignment)
            ret.append((label, getattr(exc, "formatted_output", f"{repr(exc)} (formatted_output is absent)"), True))
        return ret

    def cfg_file_names(self, device: Device) -> List[str]:
        return [f"{device.hostname}.cfg"]

    def entire_config_dest_path(self, device, config_path: str) -> str:
        """Формирует путь к конфигу в директории destname.

        Например, для устройства с hostname `my-device`:
        ```
        >>> device.entire_config_dest_path("/etc/frr/frr.conf")
        'my-device.cfg/etc/frr/frr.conf'
        >>>
        ```
        """
        # NOTE: с полученным `config_path` работаем через `posixpath`, а не через `os.path`, потому что
        #       entire-путь POSIX-специфичный; но в конце формируем путь через `os.path` для текущей платформы
        if not posixpath.abspath(config_path):
            raise RuntimeError(f"Want absolute config path, but relative received: {config_path}")
        cfg_files = self.cfg_file_names(device)
        # NOTE: получаем путь без "/" в начале, например, "etc/frr/frr.conf"
        relative_config_path = posixpath.relpath(config_path, "/")
        dest_config_path_parts = [cfg_files[0]] + relative_config_path.split(posixpath.sep)
        return os.path.join(*dest_config_path_parts)



from dataclasses import dataclass

from adaptix import Retort, name_mapping, NameStyle
from dataclass_rest import get
from dataclass_rest.client_protocol import FactoryProtocol

from .client import BaseNetboxClient


@dataclass
class Status:
    netbox_version: str
    plugins: dict[str, str]


class NetboxStatusClient(BaseNetboxClient):
    def _init_response_body_factory(self) -> FactoryProtocol:
        return Retort(recipe=[
            name_mapping(name_style=NameStyle.LOWER_KEBAB)
        ])

    @get("status")
    def status(self) -> Status:
        ...
